import warnings
from copy import copy, deepcopy
import numbers
import matplotlib.font_manager
import nibabel
import numpy as np
import yaml

import mdt
import mdt.visualization.layouts
from mdt.lib.nifti import load_nifti, NiftiInfoDecorated
from mdt.visualization.dict_conversion import StringConversion, \
    SimpleClassConversion, IntConversion, SimpleListConversion, BooleanConversion, \
    ConvertDictElements, ConvertDynamicFromModule, FloatConversion, WhiteListConversion, ConvertListElements
from mdt.visualization.layouts import Rectangular
from mdt.visualization.maps.utils import get_shortest_unique_names, find_all_nifti_files

__author__ = 'Robbert Harms'
__date__ = "2016-09-02"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ConvertibleConfig(object):
    """Base class for convertible configuration containers.

    Objects implementing this interface return a conversion specification that describes how the implementing
    class can be converted to and from a simple dictionary containing only primitives and simple
    data structures.
    """

    @classmethod
    def get_conversion_info(cls):
        """Get the conversion specification information for this class.

        Returns:
            mdt.visualization.dict_conversion.ConversionSpecification: the conversion specification
        """
        raise NotImplementedError()


class SimpleConvertibleConfig(object):
    """Offers an simplified implementation of convertible configs.

    In addition to slightly simplifying the creation of the conversion specification, it also implements
    object equality checking (__eq__ and__ne__) as this is commonly used in the configuration objects,
    and it offers a default representation (__repr__) implementation based on the conversion specification.
    """

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        """Simplifies the creation of the conversion info by returning a dictionary of conversion specifications.

        Returns:
            dict: a dictionary with for every attribute of the class a conversion specification (instance of
                :class:`~mdt.visualization.dict_conversion.ConversionSpecification`) describing how to
                transform every attribute of the class.
        """
        raise NotImplementedError()

    def __repr__(self):
        return str(self.get_conversion_info().to_dict(self))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        for key, value in self.__dict__.items():
            if value != getattr(other, key) or (value is not None and getattr(other, key) is None) \
                or (value is None and getattr(other, key) is not None):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class DataInfo(object):
    """Information about the maps we are viewing."""

    def get_map_names(self):
        """Get a list of the names of all the maps in this data info object.

        Returns:
            list: the list of map names
        """
        raise NotImplementedError()

    def get_map_data(self, map_name):
        """Get the data for the indicated map.

        Args:
            map_name (str): the name of the map we want the data of

        Returns:
            ndarray: the data of the given map
        """
        raise NotImplementedError()

    def get_single_map_info(self, map_name):
        """Get an information object for a single map.

        Args:
            map_name (str): the name of the map we want information about

        Returns:
            SingleMapInfo: information object about that map
        """
        raise NotImplementedError()

    def get_file_path(self, map_name):
        """Get the file name of the given map

        Returns:
            None if the map could not be found on dir, else a string with the file path.
        """
        raise NotImplementedError()

    def get_file_paths(self):
        """Get the file paths to each of the maps.

        If one of the maps does not have a file path, None is returned.

        Returns:
            dict: mapping map names to the file paths for each of the maps in this information container.
        """
        raise NotImplementedError()

    def get_max_dimension(self, map_names=None):
        """Get the minimum of the maximum dimension index over the maps

        Args:
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: either, 0, 1, 2 as the maximum dimension index in the maps.
        """
        raise NotImplementedError()

    def get_max_slice_index(self, dimension, map_names=None):
        """Get the maximum slice index in the given map on the given dimension.

        Args:
            dimension (int): the dimension we want the slice index of (maximum 3)
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the minimum of the maximum slice indices over the given maps in the given dimension.
        """
        raise NotImplementedError()

    def get_max_volume_index(self, map_names=None):
        """Get the maximum volume index in the given maps.

        In contrast to the max dimension and max slice index functions, this gives the maximum over all the
        images. This since handling different volumes is implemented in the viewer.

        Args:
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum volume index in the given list of maps. Starts from 0.
        """
        raise NotImplementedError()

    def get_index_first_non_zero_slice(self, dimension, map_names=None):
        """Get the index of the first non zero slice in the maps.

        Args:
            dimension (int): the dimension to search in
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the slice index with the first non zero values.
        """
        raise NotImplementedError()

    def slice_has_data(self, dimension, slice_index, map_names=None):
        """Check if at least one of the maps has non zero numbers on the given slice.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the index of the slice in the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            bool: true if at least on of the maps has data in the given slice
        """
        raise NotImplementedError()

    def get_max_x_index(self, dimension, rotate=0, map_names=None):
        """Get the maximum x index supported over the images.

        In essence this gets the lowest x index found.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the rotation factor by which we rotate the slices within the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum x-index found.
        """
        raise NotImplementedError()

    def get_max_y_index(self, dimension, rotate=0, map_names=None):
        """Get the maximum y index supported over the images.

        In essence this gets the lowest y index found.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the rotation factor by which we rotate the slices within the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum y-index found.
        """
        raise NotImplementedError()

    def get_bounding_box(self, dimension, slice_index, volume_index, rotate, map_names=None):
        """Get the bounding box of the images.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the slice index in that dimension
            volume_index (int): the current volume index
            rotate (int): the angle by which to rotate the image before getting the bounding box
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            tuple of :class:`Point2d`: two points designating the upper left corner and the lower right corner of the
                bounding box.
        """
        raise NotImplementedError()


class SimpleDataInfo(DataInfo):

    def __init__(self, maps):
        """A container for basic information about the volume maps we are viewing.

        Args:
            maps (dict): the dictionary with the maps to view, these maps can either be arrays with values,
                nibabel proxy images or SingleMapInfo objects.
        """
        self._input_maps = maps

    @classmethod
    def from_paths(cls, paths):
        """Load all the nifti files from the given paths.

        For paths that are directories we load all the elements inside that directory (but without recursion).

        Args:
            list of str: the list of paths to load (directories and files)

        Returns:
            SimpleDataInfo: the simple data info
        """
        nifti_files = find_all_nifti_files(paths)
        return cls(_load_data_info(nifti_files))

    def get_updated(self, updates=None, removals=None):
        """Get a new simple data info object that includes the given new maps.

        In the case of double map names the old maps are overwritten.

        Args:
            updates (dict): the dictionary with the maps to view, these maps can either be arrays with values,
                nibabel proxy images or SingleMapInfo objects.
            removals (list): a list of maps to remove

        Returns:
            SimpleDataInfo: a new updated data info object
        """
        new_maps = copy(self._input_maps)

        if updates:
            new_maps.update(updates)

        if removals:
            for map_name in removals:
                if map_name in new_maps:
                    del new_maps[map_name]

        return SimpleDataInfo(new_maps)

    def get_map_names(self):
        return list(self._input_maps.keys())

    def get_map_data(self, map_name):
        return self.get_single_map_info(map_name).data

    def get_single_map_info(self, map_name):
        value = self._input_maps[map_name]
        if not isinstance(value, SingleMapInfo):
            return SingleMapInfo(value)
        return value

    def get_file_path(self, map_name):
        return self.get_single_map_info(map_name).file_path

    def get_file_paths(self):
        return {map_name: self.get_file_path(map_name) for map_name in self._input_maps.keys()}

    def get_max_dimension(self, map_names=None):
        """Get the minimum of the maximum dimension index over the maps

        Args:
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: either, 0, 1, 2 as the maximum dimension index in the maps.
        """
        map_names = map_names or self._input_maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return min(self.get_single_map_info(map_name).max_dimension() for map_name in map_names)

    def get_max_slice_index(self, dimension, map_names=None):
        """Get the maximum slice index in the given map on the given dimension.

        Args:
            dimension (int): the dimension we want the slice index of (maximum 3)
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the minimum of the maximum slice indices over the given maps in the given dimension.
        """
        map_names = map_names or self._input_maps.keys()
        max_dimension = self.get_max_dimension(map_names)
        if not map_names:
            raise ValueError('No maps to search in.')
        if dimension > max_dimension:
            raise ValueError('Dimension can not exceed {}.'.format(max_dimension))
        return min(self.get_single_map_info(map_name).max_slice_index(dimension) for map_name in map_names)

    def get_max_volume_index(self, map_names=None):
        """Get the maximum volume index in the given maps.

        In contrast to the max dimension and max slice index functions, this gives the maximum over all the
        images. This since handling different volumes is implemented in the viewer.

        Args:
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum volume index in the given list of maps. Starts from 0.
        """
        map_names = map_names or self._input_maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return max(self.get_single_map_info(map_name).max_volume_index() for map_name in map_names)

    def get_index_first_non_zero_slice(self, dimension, map_names=None):
        """Get the index of the first non zero slice in the maps.

        Args:
            dimension (int): the dimension to search in
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the slice index with the first non zero values.
        """
        map_names = map_names or self._input_maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        for map_name in map_names:
            index = self.get_single_map_info(map_name).get_index_first_non_zero_slice(dimension)
            if index is not None:
                return index
        return 0

    def slice_has_data(self, dimension, slice_index, map_names=None):
        """Check if at least one of the maps has non zero numbers on the given slice.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the index of the slice in the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            bool: true if at least on of the maps has data in the given slice
        """
        map_names = map_names or self._input_maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        for map_name in map_names:
            if self.get_single_map_info(map_name).slice_has_data(dimension, slice_index):
                return True
        return False

    def get_max_x_index(self, dimension, rotate=0, map_names=None):
        """Get the maximum x index supported over the images.

        In essence this gets the lowest x index found.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the rotation factor by which we rotate the slices within the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum x-index found.
        """
        map_names = map_names or self._input_maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return min(self.get_single_map_info(map_name).get_max_x_index(dimension, rotate) for map_name in map_names)

    def get_max_y_index(self, dimension, rotate=0, map_names=None):
        """Get the maximum y index supported over the images.

        In essence this gets the lowest y index found.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the rotation factor by which we rotate the slices within the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum y-index found.
        """
        map_names = map_names or self._input_maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return min(self.get_single_map_info(map_name).get_max_y_index(dimension, rotate) for map_name in map_names)

    def get_bounding_box(self, dimension, slice_index, volume_index, rotate, map_names=None):
        """Get the bounding box of the images.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the slice index in that dimension
            volume_index (int): the current volume index
            rotate (int): the angle by which to rotate the image before getting the bounding box
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            tuple of Point: two point designating first the upper left corner and second the lower right corner of the
                bounding box.
        """
        map_names = map_names or self._input_maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        bounding_boxes = [self.get_single_map_info(map_name).get_bounding_box(dimension, slice_index, volume_index, rotate)
                          for map_name in map_names]

        p0x = min([bbox[0].x for bbox in bounding_boxes])
        p0y = min([bbox[0].y for bbox in bounding_boxes])
        p1x = max([bbox[1].x for bbox in bounding_boxes])
        p1y = max([bbox[1].y for bbox in bounding_boxes])

        return Point2d(p0x, p0y), Point2d(p1x, p1y)


class SingleMapInfo(object):

    def __init__(self, data, file_path=None):
        """Holds information about a single map.

        Args:
            data (ndarray or :class:`nibabel.spatialimages.SpatialImage`): the value of the map or the proxy to it
            file_path (str): optionally, the file path with the location of this map.
                If not set we try to retreive it from the data if the data is of subclass
                :class:`mdt.lib.nifti.NiftiInfoDecorated`.
        """
        self._data = data
        self._shape = self._data.shape
        self._file_path = file_path

        if self._file_path is None:
            if isinstance(data, NiftiInfoDecorated):
                self._file_path = data.nifti_info.filepath

    @classmethod
    def from_file(cls, nifti_path):
        return cls(load_nifti(nifti_path), nifti_path)

    @property
    def shape(self):
        return deepcopy(self._shape)

    @property
    def data(self):
        if isinstance(self._data, nibabel.spatialimages.SpatialImage):
            return self._data.get_data()
        return self._data

    @property
    def file_path(self):
        return self._file_path

    def min(self, mask=None):
        """Get the minimum value in this data.

        If a mask is provided we get the minimum value within the given mask.

        Args:
            mask (ndarray): the mask, we only include elements for which the mask > 0

        Returns:
            float: the minimum value
        """
        if mask is not None:
            return mdt.create_roi(self.data, mask).min()
        return self.data.min()

    def max(self, mask=None):
        """Get the maximum value in this data.

        If a mask is provided we get the maximum value within the given mask.

        Args:
            mask (ndarray): the mask, we only include elements for which the mask > 0

        Returns:
            float: the maximum value
        """
        if mask is not None:
            return mdt.create_roi(self.data, mask).max()
        return self.data.max()

    def min_max(self, mask=None):
        """Get the minimum and maximum value in this data.

        If a mask is provided we get the min and max value within the given mask.

        Infinities and NaN's are ignored by this algorithm.

        Args:
            mask (ndarray): the mask, we only include elements for which the mask > 0

        Returns:
            tuple: (min, max) the minimum and maximum values
        """
        if mask is not None:
            roi = mdt.create_roi(self.data, mask)
            return np.nanmin(roi), np.nanmax(roi)
        return np.nanmin(self.data), np.nanmax(self.data)

    def has_nan(self):
        """Check if this data has any NaNs in it.

        Returns:
            boolean: True if there are NaN's anywhere in the data, false otherwise.
        """
        return np.isnan(self.data).any()

    def max_dimension(self):
        """Get the maximum dimension index in this map.

        The maximum value returned by this method is 2 and the minimum is 0.

        Returns:
            int: in the range 0, 1, 2
        """
        return min(len(self.shape), 3) - 1

    def max_slice_index(self, dimension):
        """Get the maximum slice index on the given dimension.

        Args:
            dimension (int): the dimension we want the slice index of (maximum 3)

        Returns:
            int: the maximum slice index in the given dimension.
        """
        return self.shape[dimension] - 1

    def slice_has_data(self, dimension, slice_index):
        """Check if this map has non zero values in the given slice index.

        Args:
            dimension (int): the dimension we want the slice index of (maximum 3)
            slice_index (int): the slice index to look in

        Returns:
            int: the maximum slice index in the given dimension.
        """
        slice_indexing = [slice(None)] * (self.max_dimension() + 1)
        slice_indexing[dimension] = slice_index
        return np.count_nonzero(self.data[slice_indexing])

    def max_volume_index(self):
        """Get the maximum volume index in this map.

        The minimum is 0.

        Returns:
            int: the maximum volume index.
        """
        if len(self.shape) > 3:
            return self.shape[3] - 1
        return 0

    def get_index_first_non_zero_slice(self, dimension):
        """Get the index of the first non zero slice in this map.

        Args:
            dimension (int): the dimension to search in

        Returns:
            int: the slice index with the first non zero values.
        """
        slice_index = [slice(None)] * (self.max_dimension() + 1)

        if dimension > len(slice_index) - 1:
            raise ValueError('The given dimension {} is not supported.'.format(dimension))

        for index in range(self.shape[dimension]):
            slice_index[dimension] = index
            if np.count_nonzero(self.data[slice_index]) > 0:
                return index
        return 0

    def get_max_x_index(self, dimension, rotate=0):
        """Get the maximum x index.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the value by which to rotate the slices in the given dimension

        Returns:
            int: the maximum x index
        """
        shape = list(self.shape)[0:3]

        if len(shape) > 2:
            del shape[dimension]

        if rotate // 90 % 2 == 0:
            return max(0, shape[1] - 1)
        return max(0, shape[0] - 1)

    def get_max_y_index(self, dimension, rotate=0):
        """Get the maximum y index.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the value by which to rotate the slices in the given dimension

        Returns:
            int: the maximum y index
        """
        shape = list(self.shape)[0:3]

        if len(shape) > 2:
            del shape[dimension]

        if rotate // 90 % 2 == 0:
            return max(0, shape[0] - 1)
        return max(0, shape[1] - 1)

    def get_size_in_dimension(self, dimension, rotate=0):
        """Get the shape of the 2d view on the data in the given dimension.

        This basically returns a pair of (max_x, max_y).

        Args:
            dimension (int): the dimension to search in
            rotate (int): the value by which to rotate the slices in the given dimension

        Returns:
            tuple: (max_x, max_y)
        """
        return self.get_max_x_index(dimension, rotate), self.get_max_y_index(dimension, rotate)

    def get_bounding_box(self, dimension, slice_index, volume_index, rotate):
        """Get the bounding box of this map when displayed using the given indicing.

        This only works if the edges of the images are exactly zero, that is, it only works with masked datasets.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the slice index in that dimension
            volume_index (int): the current volume index
            rotate (int): the angle by which to rotate the image before getting the bounding box

        Returns:
            tuple of Point: two point designating first the upper left corner and second the lower right corner of the
                bounding box.
        """
        def bbox(image):
            rows = np.any(image, axis=1)
            cols = np.any(image, axis=0)

            rows_where = np.where(rows)

            if np.size(rows_where):
                row_min, row_max = np.where(rows)[0][[0, -1]]
                column_min, column_max = np.where(cols)[0][[0, -1]]

                return row_min, row_max, column_min, column_max
            return 0, image.shape[0]-1, 0, image.shape[1]-1

        if len(self.shape) == 2:
            image = self.data
        else:
            slice_indexing = [slice(None)] * (self.max_dimension() + 1)
            slice_indexing[dimension] = slice_index
            image = self.data[slice_indexing]

        if len(image.shape) > 2:
            if image.shape[2] > 1:
                image = image[..., volume_index]
            else:
                image = image[..., 0]

        if rotate:
            image = np.rot90(image, rotate // 90)

        row_min, row_max, column_min, column_max = bbox(image)
        return Point2d(column_min, row_min), Point2d(column_max, row_max)


class Zoom(SimpleConvertibleConfig):

    def __init__(self, p0, p1):
        """Container for zooming a map between the two given points.

        Args:
            p0 (Point2d): the lower left corner of the zoomed area
            p1 (Point2d): the upper right corner of the zoomed area
        """
        self.p0 = p0
        self.p1 = p1

        if p0.x > p1.x or p0.y > p1.y:
            raise ValueError('The lower left point ({}, {}) should be smaller than the upper right point ({}, {})'.
                             format(p0.x, p0.y, p1.x, p1.y))

        if p0.x < 0 or p0.y < 0 or p1.x < 0 or p1.y < 0:
            raise ValueError('The zoom box ({}, {}), ({}, {}) can not '
                             'be negative in any way.'.format(p0.x, p0.y, p1.x, p1.y))

        if self.p0 is None or self.p1 is None:
            raise ValueError('One of the zoom points is None.')

    @classmethod
    def from_coords(cls, x0, y0, x1, y1):
        return cls(Point2d(x0, y0), Point2d(x1, y1))

    @classmethod
    def no_zoom(cls):
        return cls(Point2d(0, 0), Point2d(0, 0))

    @classmethod
    def _get_attribute_conversions(cls):
        point_converter = Point2d.get_conversion_info()
        return {'p0': point_converter,
                'p1': point_converter}

    def get_rotated(self, rotation, x_dimension, y_dimension):
        """Return a new Zoom instance rotated with the given factor.

        This rotates the zoom box in the same way as the image is rotated.

        Args:
            rotation (int): the rotation by which to rotate in steps of 90 degrees
            x_dimension (int): the dimension of the image in the x coordinate
            y_dimension (int): the dimension of the image in the y coordinate

        Returns:
            Zoom: the rotated instance
        """
        dimensions = [x_dimension, y_dimension]
        p0 = self.p0
        p1 = self.p1

        nmr_90_rotations = rotation % 360 // 90

        for _ in range(nmr_90_rotations):
            dimensions = np.roll(dimensions, 1)

            new_p0 = Point2d(np.min([dimensions[0] - p0.y, dimensions[0] - p1.y]), np.min([p0.x, p1.x]))
            new_p1 = Point2d(np.max([dimensions[0] - p0.y, dimensions[0] - p1.y]), np.max([p0.x, p1.x]))

            p0 = new_p0
            p1 = new_p1

        if p0.x >= dimensions[0] - 1 or p0.x < 0:
            p0 = p0.get_updated(x=0)

        if p0.y >= dimensions[1] - 1 or p0.y < 0:
            p0 = p0.get_updated(y=0)

        if p1.x >= dimensions[0] - 1:
            p1 = p1.get_updated(x=dimensions[0] - 1)

        if p1.y >= dimensions[1] - 1:
            p1 = p1.get_updated(y=dimensions[1] - 1)

        return Zoom(p0, p1)

    def apply(self, data):
        """Apply the zoom to the given 2d array and return the new array.

        Args:
            data (ndarray): the data to zoom in on
        """
        correct = self.p0.x < data.shape[1] and self.p1.x < data.shape[1] \
                  and self.p0.y < data.shape[0] and self.p1.y < data.shape[0] \
                  and self.p0.x < self.p1.x and self.p0.y < self.p1.y
        if correct:
            return data[self.p0.y:self.p1.y, self.p0.x:self.p1.x]
        return data


class ColorbarSettings(SimpleConvertibleConfig):

    def __init__(self, visible=None, nmr_ticks=None, location=None, power_limits=None, round_precision=None):
        """Container for all colorbar related settings.

        Args:
            visible (boolean): if the colorbar is to be shown
            nmr_ticks (int): the number of ticks
            location (str): the location of the colorbar, one of 'right', 'left', 'top' or 'bottom'
            power_limits (tuple): size thresholds for scientific notation. The default is (-3, 4) which uses scientific
                notation for numbers less than 1e-3 or greater than 1e4.
            round_precision (int): how much digits (precision) after the decimal point.
        """
        self.nmr_ticks = nmr_ticks
        self.location = location
        self.visible = visible
        self.power_limits = power_limits
        self.round_precision = round_precision

        if self.location is not None and self.location not in ['left', 'right', 'bottom', 'top']:
            raise ValueError("The colorbar location is '{}' which is not "
                             "one of 'left', 'bottom', 'right', 'top'.".format(str(self.location)))

        if self.nmr_ticks is not None:
            self.nmr_ticks = int(self.nmr_ticks)
            if self.nmr_ticks <= 0:
                raise ValueError("The number of ticks in the colorbar needs to be a positive integer.")

        if self.visible is not None:
            self.visible = bool(self.visible)

        if self.power_limits is not None:
            if not isinstance(self.power_limits, (tuple, list)):
                raise ValueError('The power limits should be a tuple or list.')
            if len(self.power_limits) != 2:
                raise ValueError('The power limits should hold '
                                 'exactly two elements, {} given.'.format(len(self.power_limits)))
            if not all(isinstance(el, numbers.Integral) for el in self.power_limits):
                raise ValueError('The power limits should be integers.')
            if self.power_limits[0] > self.power_limits[1]:
                raise ValueError('The lower power limit should be lower than the upper power limit.')

        if self.round_precision is not None:
            self.round_precision = int(self.round_precision)

    @staticmethod
    def get_default():
        return ColorbarSettings(visible=True, nmr_ticks=4, location='right', power_limits=(-3, 4),
                                round_precision=3)

    @classmethod
    def _get_attribute_conversions(cls):
        return {
            'nmr_ticks': IntConversion(),
            'location': StringConversion(),
            'visible': BooleanConversion(),
            'power_limits': SimpleListConversion(),
            'round_precision': IntConversion()
        }

    def get_preferred(self, attr, other_settings=None):
        """Get the preferred value for the requested attribute.

        Other settings is a list of other colorbar settings object that are asked for the preferred value (in turn),
        if the value of this object is None.

        As a fallback, this will always use the default colorbar settings ``ColorbarSettings.get_default()`` as a final
        default.

        Args:
            attr (str): the attribute requested
            other_settings (list of ColorbarSettings): other settings to try.

        Returns:
            object: the value of the requested object
        """
        if getattr(self, attr) is not None:
            return getattr(self, attr)
        if not other_settings:
            return getattr(self.get_default(), attr)
        for other_setting in other_settings:
            return other_setting.get_preferred(attr, other_settings[1:])


class VoxelAnnotation(SimpleConvertibleConfig):

    valid_text_locations = ['upper left', 'top left', 'upper right',
                            'top right', 'bottom right', 'lower right',
                            'bottom left', 'lower left', 'north', 'south',
                            'east', 'west', 'top', 'bottom', 'left', 'right']

    def __init__(self, voxel_index, font_size=None, text_template=None, marker_size=1, text_location='upper left',
                 text_distance=0.05, arrow_width=1):
        """Container for all voxel highlighting settings.

        Args:
            voxel_index (tuple): a tuple with the voxel index location
            font_size (int): the size of the annotation text
            text_template (str): the text template, can use the placeholders ``{voxel_index}`` and ``{value}``.
            marker_size (float): the rectangular size of the voxel marker
            text_location (str): the location of the text. Valid items are:
                ``upper left``, ``top left``, ``upper right``, ``top right``, ``bottom right``, ``lower right``,
                ``bottom left``, ``lower left``, ``north``, ``south``, ``east``, ``west``, ``top``, ``bottom``, ``left``
                , ``right``.
            text_distance (float): the distance of the textbox to the marker in relative coordinates (0, 1).
            arrow_width (float): the width of the arrow
        """
        self.voxel_index = voxel_index
        self.font_size = font_size
        self.text_template = text_template or "{voxel_index}\n{value:.3g}"
        self.marker_size = marker_size
        self.text_location = text_location
        self.text_distance = text_distance
        self.arrow_width = arrow_width

        if text_location not in self.valid_text_locations:
            raise ValueError('The given text location "{}" is not '
                             'in the list of valid text locations: {}'.format(self.text_location,
                                                                              self.valid_text_locations))

        if not isinstance(self.marker_size, numbers.Real):
            raise ValueError('The marker size should be a real number, {} given.'.format(self.marker_size))

        if not isinstance(self.text_distance, numbers.Real):
            raise ValueError('The text distance should be a real number, {} given.'.format(self.text_distance))

        if not isinstance(self.arrow_width, numbers.Real):
            raise ValueError('The arrow width should be a real number, {} given.'.format(self.arrow_width))

        self.text_template.format(voxel_index=(0, 0, 0), value=0)  # validates the template

    @classmethod
    def _get_attribute_conversions(cls):
        return {
            'voxel_index': SimpleListConversion(),
            'font_size': IntConversion(),
            'text_template': StringConversion(allow_null=False),
            'marker_size': FloatConversion(),
            'text_location': StringConversion(allow_null=False),
            'text_distance': FloatConversion(allow_null=False),
            'arrow_width': FloatConversion(allow_null=False)
        }

    def validate(self, data_info):
        self.text_template.format(voxel_index=(0, 0, 0), value=0)

        if len(self.voxel_index) != 3:
            raise ValueError('The location of the annotation should consist of (x, y, z) '
                             'coordinates, {} given.'.format(self.voxel_index))
        for dim, pos in enumerate(self.voxel_index):
            try:
                max_pos = data_info.get_max_slice_index(dimension=dim)
            except ValueError:
                max_pos = None

            if max_pos is not None:
                if pos > max_pos:
                    raise ValueError('The location of a highlighted voxel in '
                                     'the {} dimension is larger than {}, {} given.'.format(dim, max_pos, pos))

            if pos < 0:
                raise ValueError('The location of a highlighted voxel in '
                                 'the {} dimension is negative, {} given.'.format(dim, pos))


class Point2d(SimpleConvertibleConfig):

    def __init__(self, x, y):
        """Container for a single point"""
        self.x = x
        self.y = y

    def get_updated(self, **kwargs):
        """Get a new :class:`Point2d` with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Point2d: a new scale with updated values.
        """
        new_values = dict(x=self.x, y=self.y)
        new_values.update(**kwargs)
        return Point2d(**new_values)

    @classmethod
    def _get_attribute_conversions(cls):
        return {'x': IntConversion(allow_null=False),
                'y': IntConversion(allow_null=False)}

    def rotate90(self, nmr_rotations):
        """Rotate this point around a 90 degree angle

        Args:
            nmr_rotations (int): the number of 90 degree rotations, can be negative

        Returns:
            Point2d: the rotated point
        """

        def rotate_coordinate(x, y, nmr_rotations):
            rotation_matrix = np.array([[0, -1],
                                        [1, 0]])
            rx, ry = x, y
            for rotation in range(1, nmr_rotations + 1):
                rx, ry = rotation_matrix.dot([rx, ry])
            return rx, ry

        return Point2d(*rotate_coordinate(self.x, self.y, nmr_rotations))


class Clipping(SimpleConvertibleConfig):

    def __init__(self, vmin=0, vmax=0, use_min=False, use_max=False):
        """Container for the map clipping information"""
        self.vmin = vmin
        self.vmax = vmax
        self.use_min = use_min
        self.use_max = use_max

        if use_min and use_max and vmin > vmax:
            raise ValueError('The minimum clipping ({}) can not be larger than the maximum clipping({})'.format(
                vmin, vmax))

    def apply(self, data):
        """Apply the clipping to the given 2d array and return the new array.

        Args:
           data (ndarray): the data to clip
        """
        if self.use_max or self.use_min:
            clipping_min = data.min()
            if self.use_min:
                clipping_min = self.vmin

            clipping_max = data.max()
            if self.use_max:
                clipping_max = self.vmax

            return np.clip(data, clipping_min, clipping_max)

        return data

    def visible_changes(self, old_clipping):
        """Checks if there are any visible changes between this clipping and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the clipping you wish to use is the one on which this method is called.

        Args:
            old_clipping (Clipping): the previous clipping

        Returns:
            bool: if the differences between this clipping and the other would result in visible differences.
        """
        if self.use_min != old_clipping.use_min or self.use_max != old_clipping.use_max:
            return True

        def visible_changes_in_min():
            if self.vmin == old_clipping.vmin:
                return False
            else:
                return self.use_min

        def visible_changes_in_max():
            if self.vmax == old_clipping.vmax:
                return False
            else:
                return self.use_max

        return visible_changes_in_max() or visible_changes_in_min()

    def get_updated(self, **kwargs):
        """Get a new Clipping object with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Clipping: a new scale with updated values.
        """
        new_values = dict(vmin=self.vmin, vmax=self.vmax, use_min=self.use_min, use_max=self.use_max)
        new_values.update(**kwargs)
        return Clipping(**new_values)

    @classmethod
    def _get_attribute_conversions(cls):
        return {'vmax': FloatConversion(allow_null=False),
                'vmin': FloatConversion(allow_null=False),
                'use_min': BooleanConversion(allow_null=False),
                'use_max': BooleanConversion(allow_null=False)}


class Scale(SimpleConvertibleConfig):

    def __init__(self, vmin=0, vmax=0, use_min=False, use_max=False):
        """Container the map scaling information"""
        self.vmin = vmin
        self.vmax = vmax
        self.use_min = use_min
        self.use_max = use_max

        if use_min and use_max and vmin > vmax:
            raise ValueError('The minimum scale ({}) can not be larger than the maximum scale ({})'.format(vmin, vmax))

    @classmethod
    def _get_attribute_conversions(cls):
        return {'vmax': FloatConversion(allow_null=False),
                'vmin': FloatConversion(allow_null=False),
                'use_min': BooleanConversion(allow_null=False),
                'use_max': BooleanConversion(allow_null=False)}

    def visible_changes(self, old_scale):
        """Checks if there are any visible changes between this scale and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the scale you wish to use is the one on which this method is called.

        Args:
            old_scale (Scale): the previous scale

        Returns:
            bool: if the differences between this scale and the other would result in visible differences.
        """
        if self.use_min != old_scale.use_min or self.use_max != old_scale.use_max:
            return True

        def visible_changes_in_min():
            if self.vmin == old_scale.vmin:
                return False
            else:
                return self.use_min

        def visible_changes_in_max():
            if self.vmax == old_scale.vmax:
                return False
            else:
                return self.use_max

        return visible_changes_in_max() or visible_changes_in_min()

    def get_updated(self, **kwargs):
        """Get a new Scale object with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Scale: a new scale with updated values.
        """
        new_values = dict(vmin=self.vmin, vmax=self.vmax, use_min=self.use_min, use_max=self.use_max)
        new_values.update(**kwargs)
        return Scale(**new_values)


class Font(SimpleConvertibleConfig):

    def __init__(self, family='sans-serif', size=14):
        """Information about the font to use

        Args:
            family: the name of the font to use
            size: the size of the font (> 0).
        """
        self.family = family
        self.size = size

        if family not in self.font_names():
            raise ValueError("The given font \"{}\" is not recognized.".format(family))
        if size < 1:
            raise ValueError("The size ({}) can not be smaller than 1".format(str(size)))

    def get_updated(self, **kwargs):
        """Get a new Font object with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Font: a new Font with updated values.
        """
        new_values = dict(family=self.family, size=self.size)
        new_values.update(**kwargs)
        return Font(**new_values)

    @property
    def name(self):
        return self.family

    @classmethod
    def font_names(cls):
        """Get the name of supported fonts

        Returns:
            list of str: the name of the supported fonts and font families.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            fonts = matplotlib.font_manager.get_fontconfig_fonts()
            names = []

            for font_name in fonts:
                try:
                    names.append(matplotlib.font_manager.FontProperties(fname=font_name).get_name())
                except RuntimeError:
                    pass

        return list(sorted(['sans-serif', 'serif', 'cursive', 'fantasy', 'monospace'])) + list(sorted(names))

    @classmethod
    def _get_attribute_conversions(cls):
        return {'family': StringConversion(),
                'size': IntConversion()}


class MapPlotConfig(SimpleConvertibleConfig):

    def __init__(self, dimension=2, slice_index=0, volume_index=0, rotate=90, colormap='hot', maps_to_show=None,
                 font=None, grid_layout=None, show_axis=False, show_titles=True, zoom=None,
                 map_plot_options=None, interpolation='bilinear', flipud=None, title=None,
                 title_spacing=None, mask_name=None, colorbar_settings=None, annotations=None):
        """Container for all plot related settings.

        Args:
            dimension (int): the dimension we are viewing
            slice_index (int): the slice in the dimension we are viewing
            volume_index (int): in the case of multiple volumes (4th dimension) which index we are in.
            rotate (int): the rotation factor, multiple of 90. By default we rotate 90 degrees to
                show most in-vivo datasets in a natural way.
            colormap (str): the name of the colormap to use
            maps_to_show (list of str): the names of the maps to show
            font (int): the font settings
            grid_layout (GridLayout): the layout of the grid
            show_axis (bool): if we show the axis or not
            show_titles (boolean): the global setting for enabling/disabling the plot titles
            zoom (Zoom): the zoom setting for all the plots
            map_plot_options (dict of SingleMapConfig): per map the map specific plot options
            interpolation (str): one of the available interpolations
            flipud (boolean): if True we flip the image upside down
            title (str): the title to this plot
            title_spacing (float): the spacing between the top of the plots and the title
            mask_name (str): the name of the mask to apply to the maps prior to display
            annotations (list of VoxelAnnotation): the voxel annotations
            colorbar_settings (ColorbarSettings): all colorbar related settings
        """
        super().__init__()

        default_values = self.get_default_values()

        self.dimension = dimension
        self.slice_index = slice_index
        self.volume_index = volume_index
        self.rotate = rotate
        self.colormap = colormap
        self.maps_to_show = maps_to_show or default_values['maps_to_show']
        self.zoom = zoom or default_values['zoom']
        self.font = font or default_values['font']
        self.show_axis = bool(show_axis)
        self.show_titles = bool(show_titles)
        self.grid_layout = grid_layout or default_values['grid_layout']
        self.interpolation = interpolation or default_values['bilinear']
        self.flipud = flipud
        if self.flipud is None:
            self.flipud = default_values['flipud']
        self.map_plot_options = map_plot_options or {}
        self.title = title
        self.title_spacing = title_spacing
        self.mask_name = mask_name
        self.annotations = annotations or []
        self.colorbar_settings = colorbar_settings or default_values['colorbar_settings']

        if interpolation not in self.get_available_interpolations():
            raise ValueError('The given interpolation ({}) is not supported.'.format(interpolation))

        if self.colormap not in self.get_available_colormaps():
            raise ValueError('The given colormap ({}) is not supported.'.format(self.colormap))

        if self.rotate not in [0, 90, 180, 270]:
            raise ValueError('The given rotation ({}) is not supported, use 90 '
                             'degree angles within 360.'.format(self.rotate))

        if self.dimension is None:
            raise ValueError('The dimension can not be None.')

        if self.slice_index is None:
            raise ValueError('The slice index can not be None.')

        if self.volume_index is None:
            raise ValueError('The volume index can not be None.')

        if self.rotate is None:
            raise ValueError('The rotation can not be None.')

        if self.dimension < 0:
            raise ValueError('The dimension can not be smaller than 0, {} given.'.format(self.dimension))

    @classmethod
    def get_available_interpolations(cls):
        return _get_available_interpolations()

    @classmethod
    def get_available_colormaps(cls):
        return _get_available_colormaps()

    @staticmethod
    def get_default_values():
        return {
            'dimension': 2,
            'slice_index': 0,
            'volume_index': 0,
            'rotate': 90,
            'colormap': 'hot',
            'maps_to_show': '',
            'zoom': Zoom.no_zoom(),
            'font': Font(),
            'show_axis': False,
            'show_titles': True,
            'grid_layout': Rectangular(),
            'interpolation': 'bilinear',
            'flipud': False,
            'map_plot_options': {},
            'title': None,
            'title_spacing': None,
            'mask_name': None,
            'annotations': [],
            'colorbar_settings': ColorbarSettings.get_default()
        }

    @classmethod
    def _get_attribute_conversions(cls):
        return {'dimension': IntConversion(),
                'slice_index': IntConversion(),
                'volume_index': IntConversion(),
                'rotate': IntConversion(),
                'colormap': StringConversion(),
                'maps_to_show': SimpleListConversion(),
                'zoom': Zoom.get_conversion_info(),
                'font': Font.get_conversion_info(),
                'show_axis': BooleanConversion(),
                'show_titles': BooleanConversion(),
                'map_plot_options': ConvertDictElements(SingleMapConfig.get_conversion_info()),
                'grid_layout': ConvertDynamicFromModule(mdt.visualization.layouts),
                'interpolation': WhiteListConversion(cls.get_available_interpolations(), 'bilinear'),
                'flipud': BooleanConversion(allow_null=False),
                'title': StringConversion(),
                'title_spacing': FloatConversion(),
                'mask_name': StringConversion(),
                'annotations': ConvertListElements(VoxelAnnotation.get_conversion_info()),
                'colorbar_settings': ColorbarSettings.get_conversion_info()
                }

    @classmethod
    def from_yaml(cls, text):
        return cls.get_conversion_info().from_dict(yaml.safe_load(text))

    @classmethod
    def from_dict(cls, config_dict):
        return cls.get_conversion_info().from_dict(config_dict)

    def to_dict(self, non_default_only=False):
        """Export this configuration to a dictionary

        Args:
            non_default_only (boolean): if True, we will only export those options that are not set to their default.

        Returns:
            dict: dict representation of the data
        """
        data = self.get_conversion_info().to_dict(self)
        default_values = self.get_default_values()

        if non_default_only:
            for key, value in list(data['map_plot_options'].items()):
                map_config = SingleMapConfig.get_conversion_info().from_dict(value)

                if map_config == SingleMapConfig():
                    del data['map_plot_options'][key]
                else:
                    data['map_plot_options'][key] = map_config.to_dict(non_default_only=non_default_only)

            for key, default in default_values.items():
                data_value = self._get_attribute_conversions()[key].from_dict(data[key])
                if data_value == default:
                    del data[key]

        return data

    def to_yaml(self, non_default_only=False):
        """Convert this configuration to a YAML string.

        Args:
            non_default_only (boolean): if True, we will only export those options that are not set to their default.

        Returns:
            str: a YAML representation of this configuration.
        """
        return yaml.safe_dump(self.to_dict(non_default_only=non_default_only))

    def visible_changes(self, old_config):
        """Checks if there are any visible changes between this configuration and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the configuration you wish to display is the one on which this method is called.

        Args:
            old_config (MapPlotConfig): the previous configuration

        Returns:
            bool: if the differences between this configuration and the other would result in visible differences.
        """
        def visible_difference_in_map_plot_options():
            if not len(self.map_plot_options) and len(old_config.map_plot_options):
                return True

            for key in set(self.map_plot_options.keys()):
                if key in self.maps_to_show:
                    if key not in old_config.map_plot_options:
                        return True
                    if self.map_plot_options[key].visible_changes(old_config.map_plot_options[key]):
                        return True
            return False

        if any(getattr(self, key) != getattr(old_config, key) for key in
               filter(lambda key: key != 'map_plot_options', self.__dict__)):
            return True

        return visible_difference_in_map_plot_options()

    def create_valid(self, data_info):
        """Creates a new configuration object with valid values.

        Args:
            data_info (mdt.visualization.maps.base.DataInfo): the data information

        Returns:
            MapPlotConfig: a valid map plot configuration
        """
        config = deepcopy(self)

        if data_info.get_map_names():
            max_dim = data_info.get_max_dimension()
            if max_dim < config.dimension:
                config.dimension = max_dim

        config.maps_to_show = list(filter(lambda k: k in data_info.get_map_names(), config.maps_to_show))
        if config.mask_name not in data_info.get_map_names():
            config.mask_name = None

        for name in config.maps_to_show:
            if name not in config.map_plot_options:
                config.map_plot_options[name] = SingleMapConfig()

        for name in list(config.map_plot_options):
            if name not in config.maps_to_show:
                del config.map_plot_options[name]

        return config

    def validate(self, data_info):
        """Check if this configuration is valid given the provided data.

        Args:
            data_info (mdt.visualization.maps.base.DataInfo): the data information

        Raises:
            Exception: can raise multiple sorts of exceptions if this config is not valid given the data.
        """
        if data_info.get_map_names():
            self._validate_maps_to_show(data_info)
            self._validate_dimension(data_info)

            for key in self.__dict__:
                if hasattr(self, '_validate_' + key):
                    getattr(self, '_validate_' + key)(data_info)
        return self

    def _validate_maps_to_show(self, data_info):
        if any(map(lambda k: k not in data_info.get_map_names(), self.maps_to_show)):
            raise ValueError('One or more of the given maps to show is not in the data: {}'.
                             format(set(self.maps_to_show).difference(set(data_info.get_map_names()))))

    def _validate_dimension(self, data_info):
        if self.maps_to_show:
            max_dim = data_info.get_max_dimension(map_names=self.maps_to_show)
            if self.dimension is None or self.dimension > max_dim:
                raise ValueError('The dimension ({}) can not be higher than {}.'.format(self.dimension, max_dim))

    def _validate_slice_index(self, data_info):
        if self.maps_to_show:
            max_slice_index = data_info.get_max_slice_index(self.dimension, map_names=self.maps_to_show)
            if self.slice_index is None or self.slice_index > max_slice_index or self.slice_index < 0:
                raise ValueError('The slice index ({}) can not be higher than '
                                 '{} or lower than 0.'.format(self.slice_index, max_slice_index))

    def _validate_volume_index(self, data_info):
        if self.maps_to_show:
            max_volume_index = data_info.get_max_volume_index(map_names=self.maps_to_show)
            if self.volume_index > max_volume_index or self.volume_index < 0:
                raise ValueError('The volume index ({}) can not be higher than '
                                 '{} or lower than 0.'.format(self.volume_index, max_volume_index))

    def _validate_zoom(self, data_info):
        if self.maps_to_show:
            max_x = data_info.get_max_x_index(self.dimension, self.rotate, map_names=self.maps_to_show)
            max_y = data_info.get_max_y_index(self.dimension, self.rotate, map_names=self.maps_to_show)

            if self.zoom.p1.x > max_x:
                raise ValueError('The zoom maximum x ({}) can not be larger than {}'.format(self.zoom.p1.x, max_x))

            if self.zoom.p1.y > max_y:
                raise ValueError('The zoom maximum y ({}) can not be larger than {}'.format(self.zoom.p1.y, max_y))

    def _validate_mask_name(self, data_info):
        if self.mask_name:
            if self.mask_name not in data_info.get_map_names():
                raise ValueError('The given global mask is not found in the list of maps.')

    def _validate_map_plot_options(self, data_info):
        def remove_not_in_data():
            for key in list(self.map_plot_options):
                if key not in data_info.get_map_names():
                    del self.map_plot_options[key]

        def get_validated():
            validated = {}
            for key, value in self.map_plot_options.items():
                if value is not None:
                    validated[key] = value.validate(data_info)
            self.map_plot_options = validated

        def remove_empty():
            empty_config = SingleMapConfig()
            empty_removed = {}
            for key, value in self.map_plot_options.items():
                if value != empty_config:
                    empty_removed[key] = value
            self.map_plot_options = empty_removed

        def create_placeholders():
            for map_name in self.maps_to_show:
                if map_name not in self.map_plot_options:
                    self.map_plot_options[map_name] = SingleMapConfig()

        remove_not_in_data()
        get_validated()
        remove_empty()
        create_placeholders()

    def _validate_annotations(self, data_info):
        if self.annotations is None:
            self.annotations = []

        for annotation in self.annotations:
            annotation.validate(data_info)


class SingleMapConfig(SimpleConvertibleConfig):

    def __init__(self, title=None, scale=None, clipping=None, colormap=None, colorbar_label=None,
                 show_title=None, title_spacing=None, mask_name=None, interpret_as_colormap=False,
                 colormap_weight_map=None,
                 colormap_order=None, colorbar_settings=None):
        """Creates the configuration for a single map plot.

        Args:
            title (str): the title of this plot, can contain latex using the matplotlib latex syntax
            scale (Scale): the scaling for the values in this map
            clipping (Clipping): the clipping to apply to the values prior to plotting
            colormap (str): the matplotlib colormap to use
            colorbar_label (str): the label for the colorbar
            show_title (boolean): if we want to show the title or not
            title_spacing (float): the spacing between the top of the plots and the title
            mask_name (str): the name of the mask used to mask the data prior to visualization
            interpret_as_colormap (boolean): if this is set to True and the referring map is a 4d volume with
                a vector of length 3 on the last dimension, then we can interpret this map as a colormap. This means
                that the elements of the last dimensions are used as (R, G, B) scalar values.
            colormap_weight_map (str): the name of another map to use as a scaling factor for this map. This is only
                used when ``interpret_as_colormap`` is set to True. This scales this map with the specified weight map.
            colormap_order (str): only used if ``interpret_as_colormap`` is used. This defines the order of the RGB
                components of the data. Valid strings are permutations of the letters RGB.
            colorbar_settings (ColorbarSettings): all colorbar related settings
        """
        super().__init__()

        default_values = self.get_default_values()

        self.title = title
        self.title_spacing = title_spacing
        self.scale = scale or default_values['scale']
        self.clipping = clipping or default_values['clipping']
        self.colormap = colormap
        self.colorbar_label = colorbar_label
        self.show_title = bool(show_title) if show_title is not None else default_values['show_title']
        self.mask_name = mask_name
        self.interpret_as_colormap = bool(interpret_as_colormap)
        self.colormap_weight_map = colormap_weight_map
        self.colormap_order = colormap_order
        self.colorbar_settings = colorbar_settings or default_values['colorbar_settings']

        if self.colormap is not None and self.colormap not in self.get_available_colormaps():
            raise ValueError('The given colormap ({}) is not supported.'.format(self.colormap))

        if colormap_order:
            if len(colormap_order) > 3 or not all(color in colormap_order.lower() for color in 'rgb'):
                raise ValueError('Incorrect colormap order specification, '
                                 'only permutations of "rgb" are allowed.'.format(colormap_order))

    @staticmethod
    def get_default_values():
        return {
            'title': None,
            'title_spacing': None,
            'scale': Scale(),
            'clipping': Clipping(),
            'colormap': None,
            'colorbar_label': None,
            'show_title': None,
            'mask_name': None,
            'interpret_as_colormap': False,
            'colormap_weight_map': None,
            'colormap_order': None,
            'colorbar_settings': ColorbarSettings()
        }

    @classmethod
    def _get_attribute_conversions(cls):
        return {'title': StringConversion(),
                'scale': Scale.get_conversion_info(),
                'clipping': Clipping.get_conversion_info(),
                'colormap': StringConversion(),
                'colorbar_label': StringConversion(),
                'title_spacing': FloatConversion(),
                'mask_name': StringConversion(),
                'show_title': BooleanConversion(),
                'interpret_as_colormap': BooleanConversion(allow_null=False),
                'colormap_weight_map': StringConversion(),
                'colormap_order': StringConversion(),
                'colorbar_settings': ColorbarSettings.get_conversion_info()}

    @classmethod
    def get_available_colormaps(cls):
        return _get_available_colormaps()

    @classmethod
    def from_yaml(cls, text):
        return cls.get_conversion_info().from_dict(yaml.safe_load(text))

    @classmethod
    def from_dict(cls, config_dict):
        return cls.get_conversion_info().from_dict(config_dict)

    def to_dict(self, non_default_only=False):
        """Export this configuration to a dictionary

        Args:
            non_default_only (boolean): if True, we will only export those options that are not set to their default.

        Returns:
            dict: dict representation of the data
        """
        data = self.get_conversion_info().to_dict(self)
        default_values = self.get_default_values()

        if non_default_only:
            for key, default in default_values.items():
                data_value = self._get_attribute_conversions()[key].from_dict(data[key])
                if data_value == default:
                    del data[key]

        return data

    def to_yaml(self, non_default_only=False):
        """Convert this configuration to a YAML string.

        Args:
            non_default_only (boolean): if True, we will only export those options that are not set to their default.

        Returns:
            str: a YAML representation of this configuration.
        """
        return yaml.safe_dump(self.to_dict(non_default_only=non_default_only))

    def visible_changes(self, old_config):
        """Checks if there are any visible changes between this configuration and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the configuration you wish to display is the one on which this method is called.

        Args:
            old_config (SingleMapConfig): the previous configuration

        Returns:
            bool: if the differences between this configuration and the other would result in visible differences.
        """
        def filtered_attributes():
            filtered = ['scale', 'clipping']
            return [key for key in self.__dict__ if key not in filtered]

        def visible_changes_in_scale():
            return self.scale.visible_changes(old_config.scale)

        def visible_changes_in_clipping():
            return self.clipping.visible_changes(old_config.clipping)

        if any(getattr(self, key) != getattr(old_config, key) for key in filtered_attributes()):
            return True

        return visible_changes_in_clipping() or visible_changes_in_scale()

    def validate(self, data_info):
        for key in self.__dict__:
            if hasattr(self, '_validate_' + key):
                getattr(self, '_validate_' + key)(data_info)
        return self

    def _validate_mask_name(self, data_info):
        if self.mask_name is not None and self.mask_name not in data_info.get_map_names():
            raise ValueError('The given mask name "{}" does not exist.'.format(self.mask_name))

    def _validate_colormap_weight_map(self, data_info):
        if self.colormap_weight_map is not None and self.colormap_weight_map not in data_info.get_map_names():
            raise ValueError('The given colormap weight map "{}" does not exist.'.format(self.colormap_weight_map))


def _get_available_interpolations():
    """The available interpolations for either the general map plot config or the map specifics.

    Do not call these for outside use, rather, consult the class method of the specific config you want to change.

    Returns:
        list of str: the list of available interpolations.
    """
    return ['none', 'nearest', 'bilinear', 'bicubic', 'spline16',
            'spline36', 'hanning', 'hamming', 'hermite', 'kaiser', 'quadric',
            'catrom', 'gaussian', 'bessel', 'mitchell', 'sinc', 'lanczos']


def _get_available_colormaps():
    """The available colormaps for either the general map plot config or the map specifics.

    Do not call these for outside use, rather, consult the class method of the specific config you want to change.

    Returns:
        list of str: the list of available colormaps.
    """
    return sorted(matplotlib.cm.datad)


def _load_data_info(nifti_files):
    """Load the data info for all the nifti files.

    In the case of conflicting simplified map names we name the maps to the shortest unique names.

    Returns:
        dict[str, SingleMapInfo]: the dictionary with the single map information
    """
    data_info = {}
    for ind, map_name in enumerate(get_shortest_unique_names(nifti_files)):
        data_info[map_name] = SingleMapInfo.from_file(nifti_files[ind])
    return data_info



from mdt import CascadeTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(CascadeTemplate):

    description = 'Cascade for Tensor.'
    models = ('BallStick_r1 (Cascade)',
              'Tensor')
    inits = {'Tensor': {'Tensor.theta': 'Stick0.theta',
                        'Tensor.phi': 'Stick0.phi'}}


class TensorFixed(CascadeTemplate):

    cascade_name_modifier = 'fixed'
    description = 'Cascade for Tensor with fixed angles.'
    models = ('BallStick_r1 (Cascade)',
              'Tensor')
    fixes = {'Tensor': {'Tensor.theta': 'Stick0.theta',
                        'Tensor.phi': 'Stick0.phi'}}



__author__ = 'Robbert Harms'
__date__ = '2017-10-23'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class NumDiffInfo(object):
    """Encapsulates information necessary for numerical differentiation of a parameter."""

    @property
    def max_step(self):
        """Get the maximum numerical differentiation step size for this parameter.

        Returns:
            float: the maximum step size for the parameter
        """
        raise NotImplementedError()

    @property
    def scaling_factor(self):
        """A scaling factor for the parameter to ensure that it's magnitude is close to unitary.

        This should return a value such that when a parameter is scaled with this value, the value is close to
        unitary.

        Returns:
            float: a scaling factor to use for the parameter
        """
        raise NotImplementedError()

    @property
    def use_bounds(self):
        """If we should use the boundary conditions for this parameter when calculating a derivative.

        This is typically set to True unless we are dealing with circular parameters.

        Returns:
            boolean: if we need to use the lower and upper bounds (defined in the parameter) when calculating
                the derivatives for this model.
        """
        raise NotImplementedError()


class SimpleNumDiffInfo(NumDiffInfo):

    def __init__(self, max_step=0.1, scale_factor=1, use_bounds=True, use_lower_bound=True, use_upper_bound=True):
        """A basic implementation of the numerical differentiation info for a parameter.

        Args:
            max_step (float): the numerical differentiation step size
            scale_factor (float): a scaling factor to rescale the parameter a unitary range
            use_bounds (boolean): if we need to use the boundary condition for this parameter
            use_lower_bound (boolean): if we are using bounds, if we are using the lower bound
            use_upper_bound (boolean): if we are using bounds, if we are using the upper bound
        """
        self._numdiff_step = max_step
        self._scale_factor = scale_factor
        self._use_bounds = use_bounds
        self._use_lower_bound = use_lower_bound
        self._use_upper_bound = use_upper_bound

    @property
    def max_step(self):
        return self._numdiff_step

    @property
    def scaling_factor(self):
        return self._scale_factor

    @property
    def use_bounds(self):
        return self._use_bounds

    @property
    def use_lower_bound(self):
        return self._use_lower_bound

    @property
    def use_upper_bound(self):
        return self._use_upper_bound



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class TensorApparentDiffusion(LibraryFunctionTemplate):

    description = '''
        Calculates the apparent diffusion for the Tensor model.
        
        Args:
            theta: polar angle of the first vector
            phi: azimuth angle of the first vector
            psi: rotation around the first vector, used to generate the perpendicular vectors.
            d: first eigenvalue
            dperp0: second eigenvalue
            dperp1: third eigenvalue
    '''
    return_type = 'mot_float_type'
    dependencies = ['TensorSphericalToCartesian']
    parameters = ['theta', 'phi', 'psi', 'd', 'dperp0', 'dperp1', 'g']
    cl_code = '''
        mot_float_type4 vec0, vec1, vec2;
        TensorSphericalToCartesian(theta, phi, psi, &vec0, &vec1, &vec2);
        
        return  d *      pown(dot(vec0, g), 2) +
                dperp0 * pown(dot(vec1, g), 2) +
                dperp1 * pown(dot(vec2, g), 2);
    '''



import numpy as np
from mdt import CompartmentTemplate
from mdt.lib.post_processing import DTIMeasures

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def extra_dti_results(results_dict):
    output = {}

    matrix = np.array(
        [[results_dict['D_{}{}'.format(i, j) if 'D_{}{}'.format(i, j) in results_dict else 'D_{}{}'.format(j, i)]
          for j in range(3)] for i in range(3)]).transpose()

    eigen_values, eigen_vectors = np.linalg.eig(matrix)

    output.update({'vec{}'.format(ind): eigen_vectors[:, ind] for ind in range(3)})
    output['d'] = eigen_values[:, 0]
    output['dperp0'] = eigen_values[:, 1]
    output['dperp1'] = eigen_values[:, 2]

    output.update(DTIMeasures.extra_optimization_maps(output))

    return output


class SymmetricNonParametricTensor(CompartmentTemplate):

    description = '''
        The Tensor model in which a symmetric D matrix is optimized directly,
        without vector/diffusivity parameterization.
    '''
    parameters = ('g', 'b',
                  'Tensor_D_00(D_00)', 'Tensor_D_01(D_01)', 'Tensor_D_02(D_02)',
                                       'Tensor_D_11(D_11)', 'Tensor_D_12(D_12)',
                                                            'Tensor_D_22(D_22)')
    cl_code = '''
        double diff = g.x * g.x * D_00 +
                      g.y * g.x * D_01 * 2 +
                      g.y * g.y * D_11 +
                      g.z * g.x * D_02 * 2 +
                      g.z * g.y * D_12 * 2 +
                      g.z * g.z * D_22;
        return exp(-b * diff);
    '''
    extra_optimization_maps = [extra_dti_results]



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Mask the background using the given mask.

This function multiplies a given volume (or list of volumes) with a binary mask.
"""
import argparse
import glob
import os
import mdt
from argcomplete.completers import FilesCompleter

from mdt.lib.nifti import nifti_filepath_resolution
from mdt.lib.shell_utils import BasicShellApplication
import textwrap

from mdt.utils import split_image_path

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ApplyMask(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-apply-mask data.nii.gz -m roi_mask_0_50.nii.gz
            mdt-apply-mask *.nii.gz -m my_mask.nii.gz
        ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('mask', help='the (brain) mask to use').completer = \
            FilesCompleter(['nii', 'gz', 'hdr', 'img'], directories=False)

        parser.add_argument('input_files', metavar='input_files', nargs="+", type=str,
                            help="The input images to use")

        parser.add_argument('--overwrite', dest='overwrite', action='store_true',
                            help="Overwrite the original images, if not set we create an output file.")
        parser.set_defaults(overwrite=False)

        return parser

    def run(self, args, extra_args):
        mask = mdt.load_brain_mask(nifti_filepath_resolution(os.path.realpath(args.mask)))

        file_names = []
        for file in args.input_files:
            file_names.extend(glob.glob(file))

        for file in file_names:
            if args.overwrite:
                mdt.apply_mask_to_file(file, mask)
            else:
                folder, basename, ext = split_image_path(nifti_filepath_resolution(os.path.realpath(file)))
                mdt.apply_mask_to_file(file, mask, output_fname=os.path.join(folder, basename + '_masked' + ext))


def get_doc_arg_parser():
    return ApplyMask().get_documentation_arg_parser()


if __name__ == '__main__':
    ApplyMask().start()



from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Kurtosis(CompositeModelTemplate):

    description = 'The standard Kurtosis model with in vivo defaults.'
    model_expression = '''
        S0 * KurtosisTensor
    '''
    volume_selection = {'unweighted_threshold': 25e6,
                        'use_unweighted': True,
                        'use_weighted': True,
                        'min_bval': 0,
                        'max_bval': 3e9 + 0.1e9}



import numpy as np
from mdt import CompositeModelTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CHARMED_r1(CompositeModelTemplate):

    description = 'The CHARMED model with 1 restricted compartments'

    model_expression = '''
        S0 * ( (Weight(w_hin0) * Tensor) +
               (Weight(w_res0) * CHARMEDRestricted(CHARMEDRestricted0))
               )
    '''

    inits = {'Tensor.d': 2e-9,
             'Tensor.dperp0': 1e-9,
             'Tensor.dperp1': 1e-9,
             'CHARMEDRestricted0.d': 1e-9}

    extra_optimization_maps = [
        lambda results: {'FR': results['w_res0.w']},
        lambda results: {'FR.std': results['w_res0.w.std']}
    ]
    extra_sampling_maps = [
        lambda samples: {'FR': np.mean(samples['w_res0.w'], axis=1),
                         'FR.std': np.std(samples['w_res0.w'], axis=1)}
    ]


class CHARMED_r1_Gelderen(CHARMED_r1):

    description = 'The CHARMED model using the van Gelderen cylindrical diffusion model.'
    model_expression = '''
        S0 * ( (Weight(w_hin0) * Tensor) +
               (Weight(w_res0) * CHARMEDRestrictedGelderen(CHARMEDRestricted0))
               )
    '''


class CHARMED_r2(CompositeModelTemplate):

    description = 'The CHARMED model with 2 restricted compartments'

    model_expression = '''
        S0 * ( (Weight(w_hin0) * Tensor) +
               (Weight(w_res0) * CHARMEDRestricted(CHARMEDRestricted0)) +
               (Weight(w_res1) * CHARMEDRestricted(CHARMEDRestricted1)) )
    '''

    inits = {'Tensor.d': 2e-9,
             'Tensor.dperp0': 1e-9,
             'Tensor.dperp1': 1e-9,
             'CHARMEDRestricted0.d': 1e-9,
             'CHARMEDRestricted1.d': 1e-9,
             'w_res0.w': 0.1,
             'w_res1.w': 0.1}

    sort_maps = [('w_res0', 'w_res1'), ('CHARMEDRestricted0', 'CHARMEDRestricted1')]

    extra_optimization_maps = [
        lambda results: {'FR': 1 - results['w_hin0.w']},
        lambda results: {'FR.std': results['w_hin0.w.std']}
    ]

    extra_prior = 'return w_res1.w < w_res0.w;'


class CHARMED_r3(CompositeModelTemplate):

    description = 'The standard CHARMED model with 3 restricted compartments'

    model_expression = '''
        S0 * ( (Weight(w_hin0) * Tensor) +
               (Weight(w_res0) * CHARMEDRestricted(CHARMEDRestricted0)) +
               (Weight(w_res1) * CHARMEDRestricted(CHARMEDRestricted1)) +
               (Weight(w_res2) * CHARMEDRestricted(CHARMEDRestricted2)) )
    '''

    inits = {'Tensor.d': 2e-9,
             'Tensor.dperp0': 1e-9,
             'Tensor.dperp1': 1e-9,
             'CHARMEDRestricted0.d': 1e-9,
             'CHARMEDRestricted1.d': 1e-9,
             'CHARMEDRestricted2.d': 1e-9,
             'w_res0.w': 0.1,
             'w_res1.w': 0.1,
             'w_res2.w': 0.1}

    sort_maps = [('w_res0', 'w_res1', 'w_res2'), ('CHARMEDRestricted0', 'CHARMEDRestricted1', 'CHARMEDRestricted2')]

    extra_optimization_maps = [
        lambda results: {'FR': 1 - results['w_hin0.w']},
        lambda results: {'FR.std': results['w_hin0.w.std']}
    ]

    extra_prior = 'return w_res2.w < w_res1.w && w_res1.w < w_res0.w;'




"""Routines for fitting models on multiple subjects.

The most important part of this are the batch profiles. These encapsulate information about the subjects
and about the modelling settings. Suppose you have a directory full of subjects that you want to analyze with a
few models. One way to do that is to write some scripts yourself that walk through the directory and fit
the models to the subjects. The other way would be to implement a :class:`BatchProfile` that contains details
about your directory structure and let :func:`mdt.batch_fit` fetch all the subjects for you.

Batch profiles contain a list with subject information (see :class:`SubjectInfo`) and a list of models
we wish to apply to these subjects. Furthermore each profile should support some functionality that checks
if this profile is suitable for a given directory. Using those functions the :func:`mdt.batch_fit` can try
to auto-recognize the batch profile to use based on the profile that is suitable and returns the most subjects.
"""
import glob
import logging
import numbers
import os
from textwrap import dedent
from mdt.lib.components import get_batch_profile, get_component_list
from mdt.lib.masking import create_median_otsu_brain_mask
from mdt.protocols import load_protocol, auto_load_protocol
from mdt.utils import AutoDict, load_input_data, natural_key_sort_cb
from mdt.lib.nifti import load_nifti

__author__ = 'Robbert Harms'
__date__ = "2015-08-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BatchProfile(object):

    def get_subjects(self, data_folder):
        """Get the information about all the subjects in the current folder.

        Args:
            data_folder (str): the data folder from which to load the subjects

        Returns:
            list of :class:`SubjectInfo`: the information about the found subjects
        """
        raise NotImplementedError()

    def is_suitable(self, data_folder):
        """Check if this directory can be used to use subjects from using this batch fitting profile.

        This is used for auto detecting the best batch fitting profile to use for loading
        subjects from the given base dir.

        Args:
            data_folder (str): the data folder from which to load the subjects

        Returns:
            boolean: true if this batch fitting profile can use the subjects in the current base directory,
                false otherwise.
        """
        raise NotImplementedError()


class SimpleBatchProfile(BatchProfile):

    def __init__(self, *args, **kwargs):
        """A base class for quickly implementing a batch profile.

        Implementing classes need only implement the method :meth:`_get_subjects`, then this class will handle the rest.

        Args:
            base_directory (str): the base directory from which we will load the subjects information
        """
        super().__init__()

    def get_subjects(self, data_folder):
        return self._get_subjects(data_folder)

    def is_suitable(self, data_folder):
        return len(self.get_subjects(data_folder)) > 0

    def _autoload_noise_std(self, data_folder, subject_id, file_pattern=None):
        """Try to autoload the noise standard deviation from a noise_std file.

        Args:
            data_folder (str): the data base folder
            subject_id (str): the subject for which to use the noise std.
            file_pattern (str): optionally provide a file pattern to use (uses glob)

        Returns:
            float or None: a float if a float could be loaded from a file noise_std, else nothing.
        """
        if file_pattern:
            noise_std_files = glob.glob(file_pattern)
        else:
            noise_std_files = glob.glob(os.path.join(data_folder, subject_id, 'noise_std*'))
        if len(noise_std_files):
            with open(noise_std_files[0], 'r') as f:
                return float(f.read())
        return None

    def _get_subjects(self, data_folder):
        """Get the matching subjects from the given base dir.

        This is the only function that should be implemented to get up and running.

        Args:
            data_folder (str): the data folder from which to load the subjects

        Returns:
            list of SubjectInfo: the information about the found subjects
        """
        return []

    def _get_first_existing_file(self, filenames, default=None, prepend_path=None):
        """Tries a list of filenames and returns the first filename in the list that exists.

        Args:
            filenames (iterator): the list of filenames to search for existence
            default (str): the default value returned if none of the filenames existed
            prepend_path (str): the path to optionally prepend to every file before checking existence

        Returns:
            str: the filename of the first existing file, if prepend path is set it is included.
        """
        for fname in filenames:
            if fname:
                if prepend_path:
                    fname = os.path.join(prepend_path, fname)
                if os.path.isfile(fname):
                    return fname
        return default

    def _autoload_protocol(self, path, protocols_to_try=(), bvecs_to_try=(), bvals_to_try=(), protocol_columns=None):
        prtcl_fname = self._get_first_existing_file(protocols_to_try, prepend_path=path)
        bval_fname = self._get_first_existing_file(bvals_to_try, prepend_path=path)
        bvec_fname = self._get_first_existing_file(bvecs_to_try, prepend_path=path)

        return BatchFitProtocolLoader(
            path, protocol_fname=prtcl_fname, bvec_fname=bvec_fname,
            bval_fname=bval_fname, protocol_columns=protocol_columns)


class SubjectInfo(object):

    @property
    def subject_id(self):
        """Get the ID of this subject.

        Returns:
            str: the id of this subject
        """
        raise NotImplementedError()

    @property
    def subject_base_folder(self):
        """Get the data base folder of this subject.

        Returns:
            str: the folder with the main data of this subject, this subject's home folder.
        """
        raise NotImplementedError()

    def get_input_data(self, use_gradient_deviations=False):
        """Get the input data for this subject.

        This is the data we will use during model fitting.

        Args:
            use_gradient_deviations (boolean): if we should enable the use of the gradient deviations

        Returns:
            :class:`~mdt.utils.MRIInputData`: the input data to use during model fitting
        """
        raise NotImplementedError()

    def __str__(self):
        return dedent('''
            {class_name}
                subject_id: {subject_id}
                subject_base_folder: {subject_base_folder}
                get_input_data(): <MRIInputData object>
        '''.format(class_name=self.__class__, subject_id=self.subject_id, subject_base_folder=self.subject_base_folder))


class SimpleSubjectInfo(SubjectInfo):

    def __init__(self, subject_base_folder, subject_id, dwi_fname, protocol_loader,
                 mask_fname, gradient_deviations=None, noise_std=None):
        """This class contains all the information about found subjects during batch fitting.

        It is returned by the method get_subjects() from the class BatchProfile.

        Args:
            subject_base_folder (str): the base folder of this subject
            subject_id (str): the subject id
            dwi_fname (str): the filename with path to the dwi image
            protocol_loader (ProtocolLoader): the protocol loader that can use us the protocol
            mask_fname (str): the filename of the mask to use. If None a mask is auto generated.
            gradient_deviations (str) if given, the path to the gradient deviations
            noise_std (float, ndarray, str): either None for automatic noise detection or a float with the noise STD
                to use during fitting or an ndarray with one value per voxel.
        """
        self._subject_base_folder = subject_base_folder
        self._subject_id = subject_id
        self._dwi_fname = dwi_fname
        self._protocol_loader = protocol_loader
        self._mask_fname = mask_fname
        self._gradient_deviations = gradient_deviations
        self._noise_std = noise_std

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def subject_base_folder(self):
        return self._subject_base_folder

    def get_input_data(self, use_gradient_deviations=False):
        if use_gradient_deviations:
            gradient_deviations = self._get_gradient_deviations()
        else:
            gradient_deviations = None

        return load_input_data(self._dwi_fname,
                               self._protocol_loader.get_protocol(),
                               self._get_mask(),
                               gradient_deviations=gradient_deviations,
                               noise_std=self._noise_std)

    def _get_mask(self):
        if self._mask_fname is None or not os.path.isfile(self._mask_fname):
            logger = logging.getLogger(__name__)
            logger.info('Creating a brain mask for subject {0}'.format(self.subject_id))

            protocol = self._protocol_loader.get_protocol()
            return create_median_otsu_brain_mask(self._dwi_fname, protocol)
        return self._mask_fname

    def _get_gradient_deviations(self):
        if self._gradient_deviations is not None:
            return load_nifti(self._gradient_deviations).get_data()
        return None


class BatchSubjectSelection(object):

    def get_selection(self, subject_ids):
        """Get the selection of subjects from the given list of subjects.

        Args:
            subject_ids (list of str): the list of subject ids from which we can choose which one to process

        Returns:
            list of str: the subject ids we want to use
        """
        raise NotImplementedError()

    def get_subjects(self, subjects):
        """

        Args:
            subjects (list of :class:`SubjectInfo`): the subjects loaded from the batch profile

        Returns:
            list of :class:`SubjectInfo`: the (sub)set of subjects we will actually use during the computations
        """
        raise NotImplementedError()


class AllSubjects(BatchSubjectSelection):

    def __init__(self):
        """Selects all subjects for use in the processing"""
        super().__init__()

    def get_selection(self, subject_ids):
        return subject_ids

    def get_subjects(self, subjects):
        return subjects


class SelectedSubjects(BatchSubjectSelection):

    def __init__(self, subject_ids=None, indices=None):
        """Only process the selected subjects.

        This method allows either a selection by index (unsafe for the order may change) or by subject name/ID (more
        safe in general). If ``start_from`` is given it additionally limits the list of selected subjects to include
        only those after that index.

        This essentially creates three different subsets of the given list of subjects and it will only process
        the subjects in the intersection of all those sets.

        Set any one of the options to None to ignore that option.

        Args:
            subject_ids (str or Iterable[str]): the list of names of subjects to process
            indices (int or Iterable[int]): the list of indices of subjects we wish to process
        """
        if isinstance(subject_ids, str):
            subject_ids = [subject_ids]

        if isinstance(indices, int):
            indices = [indices]

        self.subject_ids = subject_ids
        self.indices = indices

    def get_selection(self, subject_ids):
        if self.indices is None and self.subject_ids is None:
            return subject_ids

        return_ids_ind = []
        if self.indices:
            return_ids_ind = [subject for ind, subject in enumerate(subject_ids) if ind in self.indices]

        return_ids_id = []
        if self.subject_ids:
            return_ids_id = list(filter(lambda subject_id: subject_id in self.subject_ids, subject_ids))

        return list(set(return_ids_id + return_ids_ind))

    def get_subjects(self, subjects):
        selected_subjects = self.get_selection([el.subject_id for el in subjects])
        return [subject for subject in subjects if subject.subject_id in selected_subjects]


def get_subject_selection(subjects_selection):
    """Load a subject selection object from the polymorphic input.

    Args:
        subjects_selection (:class:`~mdt.lib.batch_utils.BatchSubjectSelection` or iterable): the subjects to \
            use for processing. If None, all subjects are processed. If a list is given instead of a
            :class:`~mdt.lib.batch_utils.BatchSubjectSelection` instance, we apply the following. If the elements in that
            list are string we use it as subject ids, if they are integers we use it as subject indices.

    Returns:
        mdt.lib.batch_utils.BatchSubjectSelection: a subject selection object.

    Raises:
        ValueError: if a list is given with mixed strings and integers.
    """
    if subjects_selection is None:
        return AllSubjects()

    if isinstance(subjects_selection, BatchSubjectSelection):
        return subjects_selection

    subjects_selection = list(subjects_selection)
    if all(isinstance(el, str) for el in subjects_selection):
        return SelectedSubjects(subject_ids=subjects_selection)
    elif all(isinstance(el, numbers.Integral) for el in subjects_selection):
        return SelectedSubjects(indices=subjects_selection)
    else:
        raise ValueError('Subjects selection should contain either all strings or all integers.')


class BatchFitProtocolLoader(object):

    def __init__(self, base_dir, protocol_fname=None, protocol_columns=None, bvec_fname=None, bval_fname=None):
        """A simple protocol loader for loading a protocol from a protocol file or bvec/bval files.

        This either loads the protocol file if present, or autoloads the protocol using the auto_load_protocol
        from the protocol module.
        """
        super().__init__()
        self._base_dir = base_dir
        self._protocol_fname = protocol_fname
        self._bvec_fname = bvec_fname
        self._bval_fname = bval_fname
        self._protocol_columns = protocol_columns

    def get_protocol(self):
        if self._protocol_fname and os.path.isfile(self._protocol_fname):
            return load_protocol(self._protocol_fname)

        return auto_load_protocol(self._base_dir, protocol_columns=self._protocol_columns,
                                  bvec_fname=self._bvec_fname, bval_fname=self._bval_fname)


def batch_profile_factory(batch_profile, base_directory):
    """Wrapper function for getting a batch profile.

    Args:
        batch_profile (None, string or BatchProfile): indication of the batch profile to use.
            If a string is given it is loaded from the users home folder. Else the best matching profile is returned.
        base_directory (str): the data folder we want to use the batch profile on.

    Returns:
        BatchProfile: If the given batch profile is None we return the output from get_best_batch_profile().
            If batch profile is a string we use it from the batch profiles loader. Else we return the input.
    """
    if batch_profile is None:
        return get_best_batch_profile(base_directory)
    elif isinstance(batch_profile, str):
        return get_batch_profile(batch_profile)()
    return batch_profile


def get_best_batch_profile(data_folder):
    """Get the batch profile that best matches the given directory.

    Args:
        data_folder (str): the directory for which to get the best batch profile.

    Returns:
        BatchProfile: the best matching batch profile.
    """
    profiles = [get_batch_profile(name)() for name in get_component_list('batch_profiles')]

    best_crawler = None
    best_subjects_count = 0
    for profile in profiles:
        if profile.is_suitable(data_folder):
            tmp_count = len(profile.get_subjects(data_folder))
            if tmp_count > best_subjects_count:
                best_crawler = profile
                best_subjects_count = tmp_count

    return best_crawler


def get_subject_information(data_folder, subject_ids, batch_profile=None):
    """Loads a batch profile and finds the subject with the given subject id.

    Args:
        data_folder (str): The data folder from which to load the subjects
        subject_ids (str or list of str): the subject we would like to retrieve, or a list of subject ids.
        batch_profile (:class:`~mdt.lib.batch_utils.BatchProfile` or str): the batch profile to use,
            or the name of a batch profile to use. If not given it is auto detected.

    Returns:
        Optional[mdt.lib.batch_utils.SubjectInfo, List[mdt.lib.batch_utils.SubjectInfo]]: the subject info or list of
            subject info's of the requested subjects.

    Raises:
        ValueError: if one of the subjects could not be found.
    """
    subjects = []

    def get_subjects(subject_info):
        subjects.append(subject_info)

    batch_apply(get_subjects, data_folder, batch_profile=batch_profile,
                subjects_selection=SelectedSubjects(subject_ids))

    if isinstance(subject_ids, str):
        return subjects[0]
    return subjects


def batch_apply(func, data_folder, batch_profile=None, subjects_selection=None, extra_args=None):
    """Apply a function on the subjects found in the batch profile.

    Args:
        func (callable): the function we will apply for every subject, should accept as single argument an instance of
            :class:`SubjectInfo`.
        data_folder (str): The data folder to process
        batch_profile (BatchProfile or str): the batch profile to use,
            or the name of a batch profile to use. If not given it is auto detected.
        subjects_selection (BatchSubjectSelection or Iterable[Union[str, int]]): the
            subjects to use for processing. If None, all subjects are processed. If a list is given instead of a
            :class:`~mdt.lib.batch_utils.BatchSubjectSelection` instance, we apply the following. If the elements in that
            list are string we use it as subject ids, if they are integers we use it as subject indices.
        extra_args (list): a list of additional arguments that are passed to the function. If this is set,
            the callback function must accept these additional args.

    Returns:
        dict: per subject id the output from the function
    """
    batch_profile = batch_profile_factory(batch_profile, data_folder)
    subjects_selection = get_subject_selection(subjects_selection)

    if batch_profile is None:
        raise RuntimeError('No suitable batch profile could be '
                           'found for the directory {0}'.format(os.path.abspath(data_folder)))

    subjects = subjects_selection.get_subjects(batch_profile.get_subjects(data_folder))

    results = {}
    for subject in subjects:
        def f(subject):
            if extra_args:
                return func(subject, *extra_args)
            return func(subject)

        results[subject.subject_id] = f(subject)
    return results


class BatchFitSubjectOutputInfo(object):

    def __init__(self, output_path, subject_id, model_name):
        """This class is used in conjunction with the function :func:`run_function_on_batch_fit_output`.

        Args:
            output_path (str): the full path to the directory with the maps
            subject_id (str): the id of the current subject
            model_name (str): the name of the model (not a path)
        """
        self.subject_id = subject_id
        self.output_path = output_path
        self.model_name = model_name

    def __repr__(self):
        return dedent('''
            subject_id: {subject_id}
            output_path: {output_path}
            model_name: {model_name}
        '''.format(subject_id=self.subject_id, output_path=self.output_path, model_name=self.model_name)).strip()


def run_function_on_batch_fit_output(func, output_folder, subjects_selection=None):
    """Run a function on the output of a batch fitting routine.

    This enables you to run a function on every model output from every subject. This expects the output directory
    to contain directories and files like <subject_id>/<model_name>/<map_name>.nii.gz

    Args:
        func (Callable[[BatchFitSubjectOutputInfo], any]): the python function we should call for every map and model.
            This should accept as single parameter a :class:`BatchFitSubjectOutputInfo`.
        output_folder (str): The data input folder
        subjects_selection (BatchSubjectSelection or iterable): the subjects to use for processing.
            If None all subjects are processed.

    Returns:
        dict: indexed by subject->model_name, values are the return values of the users function
    """
    subject_ids = list(os.listdir(output_folder))
    subject_ids.sort(key=natural_key_sort_cb)

    subjects_selection = get_subject_selection(subjects_selection)
    if subjects_selection:
        subject_ids = subjects_selection.get_selection(subject_ids)

    results = AutoDict()
    for subject_id in subject_ids:
        for model_name in os.listdir(os.path.join(output_folder, subject_id)):
            info = BatchFitSubjectOutputInfo(os.path.join(output_folder, subject_id, model_name),
                                             subject_id, model_name)
            results[subject_id][model_name] = func(info)
    return results.to_normal_dict()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'TabGeneral.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_TabGeneral(object):
    def setupUi(self, TabGeneral):
        TabGeneral.setObjectName("TabGeneral")
        TabGeneral.resize(963, 704)
        self.gridLayout = QtWidgets.QGridLayout(TabGeneral)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea = QtWidgets.QScrollArea(TabGeneral)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, -289, 947, 1075))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_6.setContentsMargins(6, 6, 6, 6)
        self.gridLayout_6.setHorizontalSpacing(0)
        self.gridLayout_6.setVerticalSpacing(10)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.frame = CollapsablePanel(self.scrollAreaWidgetContents)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame.setObjectName("frame")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_13 = CollapsablePanelHeader(self.frame)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_13.setFont(font)
        self.label_13.setObjectName("label_13")
        self.verticalLayout_4.addWidget(self.label_13)
        self.line_2 = QtWidgets.QFrame(self.frame)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout_4.addWidget(self.line_2)
        self.frame_3 = CollapsablePanelContent(self.frame)
        self.frame_3.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_3.setObjectName("frame_3")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.frame_3)
        self.gridLayout_2.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_2.setHorizontalSpacing(6)
        self.gridLayout_2.setVerticalSpacing(3)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setContentsMargins(3, -1, -1, -1)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.label_10 = QtWidgets.QLabel(self.frame_3)
        self.label_10.setObjectName("label_10")
        self.horizontalLayout_7.addWidget(self.label_10)
        self.maximumDimension = QtWidgets.QLabel(self.frame_3)
        self.maximumDimension.setObjectName("maximumDimension")
        self.horizontalLayout_7.addWidget(self.maximumDimension)
        self.gridLayout_2.addLayout(self.horizontalLayout_7, 0, 2, 1, 1)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setContentsMargins(3, -1, -1, -1)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_9 = QtWidgets.QLabel(self.frame_3)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_6.addWidget(self.label_9)
        self.maximumIndex = QtWidgets.QLabel(self.frame_3)
        self.maximumIndex.setObjectName("maximumIndex")
        self.horizontalLayout_6.addWidget(self.maximumIndex)
        self.gridLayout_2.addLayout(self.horizontalLayout_6, 1, 2, 1, 1)
        self.label_14 = QtWidgets.QLabel(self.frame_3)
        self.label_14.setObjectName("label_14")
        self.gridLayout_2.addWidget(self.label_14, 0, 0, 1, 1)
        self.general_dimension = QtWidgets.QSpinBox(self.frame_3)
        self.general_dimension.setObjectName("general_dimension")
        self.gridLayout_2.addWidget(self.general_dimension, 0, 1, 1, 1)
        self.label_16 = QtWidgets.QLabel(self.frame_3)
        self.label_16.setObjectName("label_16")
        self.gridLayout_2.addWidget(self.label_16, 1, 0, 1, 1)
        self.general_slice_index = QtWidgets.QSpinBox(self.frame_3)
        self.general_slice_index.setObjectName("general_slice_index")
        self.gridLayout_2.addWidget(self.general_slice_index, 1, 1, 1, 1)
        self.label_15 = QtWidgets.QLabel(self.frame_3)
        self.label_15.setObjectName("label_15")
        self.gridLayout_2.addWidget(self.label_15, 2, 0, 1, 1)
        self.general_volume_index = QtWidgets.QSpinBox(self.frame_3)
        self.general_volume_index.setObjectName("general_volume_index")
        self.gridLayout_2.addWidget(self.general_volume_index, 2, 1, 1, 1)
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_10.setContentsMargins(3, -1, -1, -1)
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.label_25 = QtWidgets.QLabel(self.frame_3)
        self.label_25.setObjectName("label_25")
        self.horizontalLayout_10.addWidget(self.label_25)
        self.maximumVolume = QtWidgets.QLabel(self.frame_3)
        self.maximumVolume.setObjectName("maximumVolume")
        self.horizontalLayout_10.addWidget(self.maximumVolume)
        self.gridLayout_2.addLayout(self.horizontalLayout_10, 2, 2, 1, 1)
        self.gridLayout_2.setColumnStretch(1, 1)
        self.verticalLayout_4.addWidget(self.frame_3)
        self.gridLayout_6.addWidget(self.frame, 0, 0, 1, 1)
        self.general_Miscellaneous = CollapsablePanel(self.scrollAreaWidgetContents)
        self.general_Miscellaneous.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.general_Miscellaneous.setFrameShadow(QtWidgets.QFrame.Plain)
        self.general_Miscellaneous.setObjectName("general_Miscellaneous")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.general_Miscellaneous)
        self.verticalLayout_8.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_8.setSpacing(0)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.label_23 = CollapsablePanelHeader(self.general_Miscellaneous)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_23.setFont(font)
        self.label_23.setObjectName("label_23")
        self.verticalLayout_8.addWidget(self.label_23)
        self.line_6 = QtWidgets.QFrame(self.general_Miscellaneous)
        self.line_6.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_6.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_6.setObjectName("line_6")
        self.verticalLayout_8.addWidget(self.line_6)
        self.frame_10 = CollapsablePanelContent(self.general_Miscellaneous)
        self.frame_10.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_10.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_10.setObjectName("frame_10")
        self.formLayout = QtWidgets.QFormLayout(self.frame_10)
        self.formLayout.setContentsMargins(6, 6, 0, 0)
        self.formLayout.setHorizontalSpacing(6)
        self.formLayout.setVerticalSpacing(3)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(self.frame_10)
        self.label.setObjectName("label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label)
        self.general_colormap = QtWidgets.QComboBox(self.frame_10)
        self.general_colormap.setObjectName("general_colormap")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.general_colormap)
        self.label_2 = QtWidgets.QLabel(self.frame_10)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.general_rotate = QtWidgets.QComboBox(self.frame_10)
        self.general_rotate.setObjectName("general_rotate")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.general_rotate)
        self.label_4 = QtWidgets.QLabel(self.frame_10)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.general_show_axis = QtWidgets.QCheckBox(self.frame_10)
        self.general_show_axis.setText("")
        self.general_show_axis.setObjectName("general_show_axis")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.general_show_axis)
        self.label_11 = QtWidgets.QLabel(self.frame_10)
        self.label_11.setObjectName("label_11")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_11)
        self.general_flipud = QtWidgets.QCheckBox(self.frame_10)
        self.general_flipud.setText("")
        self.general_flipud.setObjectName("general_flipud")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.general_flipud)
        self.label_27 = QtWidgets.QLabel(self.frame_10)
        self.label_27.setObjectName("label_27")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.label_27)
        self.general_interpolation = QtWidgets.QComboBox(self.frame_10)
        self.general_interpolation.setObjectName("general_interpolation")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.general_interpolation)
        self.label_28 = QtWidgets.QLabel(self.frame_10)
        self.label_28.setObjectName("label_28")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_28)
        self.plot_title = QtWidgets.QLineEdit(self.frame_10)
        self.plot_title.setObjectName("plot_title")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.plot_title)
        self.label_29 = QtWidgets.QLabel(self.frame_10)
        self.label_29.setObjectName("label_29")
        self.formLayout.setWidget(8, QtWidgets.QFormLayout.LabelRole, self.label_29)
        self.mask_name = QtWidgets.QComboBox(self.frame_10)
        self.mask_name.setObjectName("mask_name")
        self.formLayout.setWidget(8, QtWidgets.QFormLayout.FieldRole, self.mask_name)
        self.label_33 = QtWidgets.QLabel(self.frame_10)
        self.label_33.setObjectName("label_33")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_33)
        self.general_show_plot_titles = QtWidgets.QCheckBox(self.frame_10)
        self.general_show_plot_titles.setText("")
        self.general_show_plot_titles.setObjectName("general_show_plot_titles")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.general_show_plot_titles)
        self.verticalLayout_8.addWidget(self.frame_10)
        self.gridLayout_6.addWidget(self.general_Miscellaneous, 8, 0, 1, 1)
        self.general_DisplayOrder = CollapsablePanel(self.scrollAreaWidgetContents)
        self.general_DisplayOrder.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.general_DisplayOrder.setFrameShadow(QtWidgets.QFrame.Plain)
        self.general_DisplayOrder.setObjectName("general_DisplayOrder")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.general_DisplayOrder)
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_7.setSpacing(0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.label_22 = CollapsablePanelHeader(self.general_DisplayOrder)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_22.setFont(font)
        self.label_22.setObjectName("label_22")
        self.verticalLayout_7.addWidget(self.label_22)
        self.line_5 = QtWidgets.QFrame(self.general_DisplayOrder)
        self.line_5.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.verticalLayout_7.addWidget(self.line_5)
        self.frame_9 = CollapsablePanelContent(self.general_DisplayOrder)
        self.frame_9.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_9.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_9.setObjectName("frame_9")
        self.gridLayout_11 = QtWidgets.QGridLayout(self.frame_9)
        self.gridLayout_11.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_11.setHorizontalSpacing(6)
        self.gridLayout_11.setVerticalSpacing(3)
        self.gridLayout_11.setObjectName("gridLayout_11")
        self.general_display_order = MapsReorderer(self.frame_9)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.general_display_order.sizePolicy().hasHeightForWidth())
        self.general_display_order.setSizePolicy(sizePolicy)
        self.general_display_order.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.general_display_order.setObjectName("general_display_order")
        self.gridLayout_11.addWidget(self.general_display_order, 0, 0, 1, 1)
        self.verticalLayout_7.addWidget(self.frame_9)
        self.gridLayout_6.addWidget(self.general_DisplayOrder, 4, 0, 1, 1)
        self.frame_6 = CollapsablePanel(self.scrollAreaWidgetContents)
        self.frame_6.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_6.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_6.setObjectName("frame_6")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.frame_6)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setSpacing(0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_21 = CollapsablePanelHeader(self.frame_6)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_21.setFont(font)
        self.label_21.setObjectName("label_21")
        self.verticalLayout_6.addWidget(self.label_21)
        self.line_4 = QtWidgets.QFrame(self.frame_6)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.verticalLayout_6.addWidget(self.line_4)
        self.frame_7 = CollapsablePanelContent(self.frame_6)
        self.frame_7.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_7.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_7.setObjectName("frame_7")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.frame_7)
        self.gridLayout_5.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_5.setHorizontalSpacing(6)
        self.gridLayout_5.setVerticalSpacing(3)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.general_map_selection = QtWidgets.QListWidget(self.frame_7)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.general_map_selection.sizePolicy().hasHeightForWidth())
        self.general_map_selection.setSizePolicy(sizePolicy)
        self.general_map_selection.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.general_map_selection.setObjectName("general_map_selection")
        self.gridLayout_5.addWidget(self.general_map_selection, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, -1, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.general_deselect_all_maps = QtWidgets.QPushButton(self.frame_7)
        self.general_deselect_all_maps.setObjectName("general_deselect_all_maps")
        self.horizontalLayout.addWidget(self.general_deselect_all_maps)
        self.general_invert_map_selection = QtWidgets.QPushButton(self.frame_7)
        self.general_invert_map_selection.setObjectName("general_invert_map_selection")
        self.horizontalLayout.addWidget(self.general_invert_map_selection)
        self.gridLayout_5.addLayout(self.horizontalLayout, 1, 0, 1, 1)
        self.verticalLayout_6.addWidget(self.frame_7)
        self.gridLayout_6.addWidget(self.frame_6, 2, 0, 1, 1)
        self.general_Info = CollapsablePanel(self.scrollAreaWidgetContents)
        self.general_Info.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.general_Info.setFrameShadow(QtWidgets.QFrame.Plain)
        self.general_Info.setObjectName("general_Info")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.general_Info)
        self.verticalLayout_9.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_9.setSpacing(0)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.label_24 = CollapsablePanelHeader(self.general_Info)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_24.setFont(font)
        self.label_24.setObjectName("label_24")
        self.verticalLayout_9.addWidget(self.label_24)
        self.line_7 = QtWidgets.QFrame(self.general_Info)
        self.line_7.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_7.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_7.setObjectName("line_7")
        self.verticalLayout_9.addWidget(self.line_7)
        self.frame_8 = CollapsablePanelContent(self.general_Info)
        self.frame_8.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_8.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_8.setObjectName("frame_8")
        self.gridLayout_12 = QtWidgets.QGridLayout(self.frame_8)
        self.gridLayout_12.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_12.setHorizontalSpacing(6)
        self.gridLayout_12.setVerticalSpacing(3)
        self.gridLayout_12.setObjectName("gridLayout_12")
        self.label_7 = QtWidgets.QLabel(self.frame_8)
        self.label_7.setObjectName("label_7")
        self.gridLayout_12.addWidget(self.label_7, 1, 0, 1, 1)
        self.general_info_nmr_maps = QtWidgets.QLabel(self.frame_8)
        self.general_info_nmr_maps.setObjectName("general_info_nmr_maps")
        self.gridLayout_12.addWidget(self.general_info_nmr_maps, 1, 1, 1, 1)
        self.verticalLayout_9.addWidget(self.frame_8)
        self.gridLayout_6.addWidget(self.general_Info, 10, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_6.addItem(spacerItem, 11, 0, 1, 1)
        self.general_Font = CollapsablePanel(self.scrollAreaWidgetContents)
        self.general_Font.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.general_Font.setFrameShadow(QtWidgets.QFrame.Plain)
        self.general_Font.setObjectName("general_Font")
        self.verticalLayout_10 = QtWidgets.QVBoxLayout(self.general_Font)
        self.verticalLayout_10.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_10.setSpacing(0)
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.label_26 = CollapsablePanelHeader(self.general_Font)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_26.setFont(font)
        self.label_26.setObjectName("label_26")
        self.verticalLayout_10.addWidget(self.label_26)
        self.line_8 = QtWidgets.QFrame(self.general_Font)
        self.line_8.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_8.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_8.setObjectName("line_8")
        self.verticalLayout_10.addWidget(self.line_8)
        self.frame_11 = CollapsablePanelContent(self.general_Font)
        self.frame_11.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_11.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_11.setObjectName("frame_11")
        self.formLayout_2 = QtWidgets.QFormLayout(self.frame_11)
        self.formLayout_2.setContentsMargins(6, 6, 0, 0)
        self.formLayout_2.setHorizontalSpacing(6)
        self.formLayout_2.setVerticalSpacing(3)
        self.formLayout_2.setObjectName("formLayout_2")
        self.label_6 = QtWidgets.QLabel(self.frame_11)
        self.label_6.setObjectName("label_6")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.general_font_family = QtWidgets.QComboBox(self.frame_11)
        self.general_font_family.setObjectName("general_font_family")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.general_font_family)
        self.label_8 = QtWidgets.QLabel(self.frame_11)
        self.label_8.setObjectName("label_8")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_8)
        self.general_font_size = QtWidgets.QSpinBox(self.frame_11)
        self.general_font_size.setMinimum(1)
        self.general_font_size.setObjectName("general_font_size")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.general_font_size)
        self.verticalLayout_10.addWidget(self.frame_11)
        self.gridLayout_6.addWidget(self.general_Font, 6, 0, 1, 1)
        self.general_Zoom = CollapsablePanel(self.scrollAreaWidgetContents)
        self.general_Zoom.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.general_Zoom.setFrameShadow(QtWidgets.QFrame.Plain)
        self.general_Zoom.setObjectName("general_Zoom")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.general_Zoom)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_17 = CollapsablePanelHeader(self.general_Zoom)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_17.setFont(font)
        self.label_17.setObjectName("label_17")
        self.verticalLayout_5.addWidget(self.label_17)
        self.line_3 = QtWidgets.QFrame(self.general_Zoom)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout_5.addWidget(self.line_3)
        self.frame_5 = CollapsablePanelContent(self.general_Zoom)
        self.frame_5.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_5.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_5.setObjectName("frame_5")
        self.gridLayout_10 = QtWidgets.QGridLayout(self.frame_5)
        self.gridLayout_10.setContentsMargins(6, 6, 0, 0)
        self.gridLayout_10.setHorizontalSpacing(6)
        self.gridLayout_10.setVerticalSpacing(3)
        self.gridLayout_10.setObjectName("gridLayout_10")
        self.general_zoom_x_1 = QtWidgets.QSpinBox(self.frame_5)
        self.general_zoom_x_1.setObjectName("general_zoom_x_1")
        self.gridLayout_10.addWidget(self.general_zoom_x_1, 1, 3, 1, 1)
        self.label_19 = QtWidgets.QLabel(self.frame_5)
        self.label_19.setObjectName("label_19")
        self.gridLayout_10.addWidget(self.label_19, 7, 0, 1, 1)
        self.label_20 = QtWidgets.QLabel(self.frame_5)
        self.label_20.setObjectName("label_20")
        self.gridLayout_10.addWidget(self.label_20, 1, 0, 1, 1)
        self.general_zoom_x_0 = QtWidgets.QSpinBox(self.frame_5)
        self.general_zoom_x_0.setObjectName("general_zoom_x_0")
        self.gridLayout_10.addWidget(self.general_zoom_x_0, 1, 1, 1, 1)
        self.general_zoom_y_1 = QtWidgets.QSpinBox(self.frame_5)
        self.general_zoom_y_1.setObjectName("general_zoom_y_1")
        self.gridLayout_10.addWidget(self.general_zoom_y_1, 7, 3, 1, 1)
        self.label_18 = QtWidgets.QLabel(self.frame_5)
        self.label_18.setObjectName("label_18")
        self.gridLayout_10.addWidget(self.label_18, 7, 2, 1, 1)
        self.general_zoom_y_0 = QtWidgets.QSpinBox(self.frame_5)
        self.general_zoom_y_0.setObjectName("general_zoom_y_0")
        self.gridLayout_10.addWidget(self.general_zoom_y_0, 7, 1, 1, 1)
        self.label_12 = QtWidgets.QLabel(self.frame_5)
        self.label_12.setObjectName("label_12")
        self.gridLayout_10.addWidget(self.label_12, 1, 2, 1, 1)
        self.general_zoom_fit = QtWidgets.QPushButton(self.frame_5)
        self.general_zoom_fit.setObjectName("general_zoom_fit")
        self.gridLayout_10.addWidget(self.general_zoom_fit, 8, 1, 1, 1)
        self.general_zoom_reset = QtWidgets.QPushButton(self.frame_5)
        self.general_zoom_reset.setObjectName("general_zoom_reset")
        self.gridLayout_10.addWidget(self.general_zoom_reset, 8, 3, 1, 1)
        self.gridLayout_10.setColumnStretch(1, 1)
        self.gridLayout_10.setColumnStretch(3, 1)
        self.verticalLayout_5.addWidget(self.frame_5)
        self.gridLayout_6.addWidget(self.general_Zoom, 5, 0, 1, 1)
        self.general_Colorbar = CollapsablePanel(self.scrollAreaWidgetContents)
        self.general_Colorbar.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.general_Colorbar.setFrameShadow(QtWidgets.QFrame.Plain)
        self.general_Colorbar.setObjectName("general_Colorbar")
        self.verticalLayout_11 = QtWidgets.QVBoxLayout(self.general_Colorbar)
        self.verticalLayout_11.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_11.setSpacing(0)
        self.verticalLayout_11.setObjectName("verticalLayout_11")
        self.label_30 = CollapsablePanelHeader(self.general_Colorbar)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_30.setFont(font)
        self.label_30.setObjectName("label_30")
        self.verticalLayout_11.addWidget(self.label_30)
        self.line_9 = QtWidgets.QFrame(self.general_Colorbar)
        self.line_9.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_9.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_9.setObjectName("line_9")
        self.verticalLayout_11.addWidget(self.line_9)
        self.frame_12 = CollapsablePanelContent(self.general_Colorbar)
        self.frame_12.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_12.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_12.setObjectName("frame_12")
        self.formLayout_3 = QtWidgets.QFormLayout(self.frame_12)
        self.formLayout_3.setContentsMargins(6, 6, 0, 0)
        self.formLayout_3.setHorizontalSpacing(6)
        self.formLayout_3.setVerticalSpacing(3)
        self.formLayout_3.setObjectName("formLayout_3")
        self.label_5 = QtWidgets.QLabel(self.frame_12)
        self.label_5.setObjectName("label_5")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.general_colorbar_nmr_ticks = QtWidgets.QSpinBox(self.frame_12)
        self.general_colorbar_nmr_ticks.setMinimum(2)
        self.general_colorbar_nmr_ticks.setProperty("value", 10)
        self.general_colorbar_nmr_ticks.setObjectName("general_colorbar_nmr_ticks")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.general_colorbar_nmr_ticks)
        self.label_31 = QtWidgets.QLabel(self.frame_12)
        self.label_31.setObjectName("label_31")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_31)
        self.general_show_colorbar = QtWidgets.QCheckBox(self.frame_12)
        self.general_show_colorbar.setText("")
        self.general_show_colorbar.setObjectName("general_show_colorbar")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.general_show_colorbar)
        self.label_32 = QtWidgets.QLabel(self.frame_12)
        self.label_32.setObjectName("label_32")
        self.formLayout_3.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_32)
        self.general_colorbar_location = QtWidgets.QComboBox(self.frame_12)
        self.general_colorbar_location.setObjectName("general_colorbar_location")
        self.general_colorbar_location.addItem("")
        self.general_colorbar_location.addItem("")
        self.general_colorbar_location.addItem("")
        self.general_colorbar_location.addItem("")
        self.formLayout_3.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.general_colorbar_location)
        self.verticalLayout_11.addWidget(self.frame_12)
        self.gridLayout_6.addWidget(self.general_Colorbar, 7, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 0, 0, 1, 1)

        self.retranslateUi(TabGeneral)
        QtCore.QMetaObject.connectSlotsByName(TabGeneral)
        TabGeneral.setTabOrder(self.scrollArea, self.general_dimension)
        TabGeneral.setTabOrder(self.general_dimension, self.general_slice_index)
        TabGeneral.setTabOrder(self.general_slice_index, self.general_volume_index)
        TabGeneral.setTabOrder(self.general_volume_index, self.general_zoom_x_0)
        TabGeneral.setTabOrder(self.general_zoom_x_0, self.general_zoom_y_0)
        TabGeneral.setTabOrder(self.general_zoom_y_0, self.general_zoom_x_1)
        TabGeneral.setTabOrder(self.general_zoom_x_1, self.general_zoom_y_1)
        TabGeneral.setTabOrder(self.general_zoom_y_1, self.general_map_selection)
        TabGeneral.setTabOrder(self.general_map_selection, self.general_display_order)
        TabGeneral.setTabOrder(self.general_display_order, self.general_colormap)
        TabGeneral.setTabOrder(self.general_colormap, self.general_rotate)
        TabGeneral.setTabOrder(self.general_rotate, self.general_show_axis)

    def retranslateUi(self, TabGeneral):
        _translate = QtCore.QCoreApplication.translate
        TabGeneral.setWindowTitle(_translate("TabGeneral", "Form"))
        self.label_13.setText(_translate("TabGeneral", "Index"))
        self.label_10.setText(_translate("TabGeneral", "/ "))
        self.maximumDimension.setText(_translate("TabGeneral", "x"))
        self.label_9.setText(_translate("TabGeneral", "/ "))
        self.maximumIndex.setText(_translate("TabGeneral", "x"))
        self.label_14.setText(_translate("TabGeneral", "Dimension:"))
        self.label_16.setText(_translate("TabGeneral", "Slice index:"))
        self.label_15.setText(_translate("TabGeneral", "Volume:"))
        self.label_25.setText(_translate("TabGeneral", "/ "))
        self.maximumVolume.setText(_translate("TabGeneral", "x"))
        self.label_23.setText(_translate("TabGeneral", "Miscellaneous"))
        self.label.setText(_translate("TabGeneral", "Colormap:"))
        self.label_2.setText(_translate("TabGeneral", "Rotate:"))
        self.label_4.setText(_translate("TabGeneral", "Show axis:"))
        self.label_11.setText(_translate("TabGeneral", "Flip up/down:"))
        self.label_27.setText(_translate("TabGeneral", "Interpolation:"))
        self.label_28.setText(_translate("TabGeneral", "Plot title:"))
        self.label_29.setText(_translate("TabGeneral", "Mask:"))
        self.label_33.setText(_translate("TabGeneral", "Show plot titles:"))
        self.label_22.setText(_translate("TabGeneral", "Display order"))
        self.label_21.setText(_translate("TabGeneral", "Map selection"))
        self.general_deselect_all_maps.setText(_translate("TabGeneral", "Deselect all"))
        self.general_invert_map_selection.setText(_translate("TabGeneral", "Invert selection"))
        self.label_24.setText(_translate("TabGeneral", "Info"))
        self.label_7.setText(_translate("TabGeneral", "Map count:"))
        self.general_info_nmr_maps.setText(_translate("TabGeneral", "-"))
        self.label_26.setText(_translate("TabGeneral", "Font"))
        self.label_6.setText(_translate("TabGeneral", "Family:"))
        self.label_8.setText(_translate("TabGeneral", "Size:"))
        self.label_17.setText(_translate("TabGeneral", "Zoom"))
        self.label_19.setText(_translate("TabGeneral", "y<sub>0</sub>:"))
        self.label_20.setText(_translate("TabGeneral", "<html><head/><body><p>x<span style=\" vertical-align:sub;\">0</span>:</p></body></html>"))
        self.label_18.setText(_translate("TabGeneral", "y<sub>1</sub>:"))
        self.label_12.setText(_translate("TabGeneral", "x<sub>1</sub>:"))
        self.general_zoom_fit.setText(_translate("TabGeneral", "Zoom fit"))
        self.general_zoom_reset.setText(_translate("TabGeneral", "Reset"))
        self.label_30.setText(_translate("TabGeneral", "Colorbar"))
        self.label_5.setText(_translate("TabGeneral", "Number of ticks:"))
        self.label_31.setText(_translate("TabGeneral", "Show colorbar:"))
        self.label_32.setText(_translate("TabGeneral", "Location:"))
        self.general_colorbar_location.setItemText(0, _translate("TabGeneral", "Right"))
        self.general_colorbar_location.setItemText(1, _translate("TabGeneral", "Bottom"))
        self.general_colorbar_location.setItemText(2, _translate("TabGeneral", "Left"))
        self.general_colorbar_location.setItemText(3, _translate("TabGeneral", "Top"))

from ..widgets import CollapsablePanel, CollapsablePanelContent, CollapsablePanelHeader, MapsReorderer



from mdt import CompositeModelTemplate
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallRacket_r1(CompositeModelTemplate):

    description = 'The default Ball & Rackets model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_res0) * Racket(Racket0)) )
    '''
    fixes = {'Racket0.d': 'Ball.d'}

    extra_optimization_maps = [
        lambda results: {'FS': results['w_res0.w']},
        lambda results: {'FS.std': results['w_res0.w.std']}
    ]
    extra_sampling_maps = [
        lambda samples: {'FS': np.mean(samples['w_res0.w'], axis=1),
                         'FS.std': np.std(samples['w_res0.w'], axis=1)}
    ]


class BallRacket_r2(CompositeModelTemplate):

    description = 'The Ball & 2x Racket model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_res0) * Racket(Racket0)) +
               (Weight(w_res1) * Racket(Racket1)) )
    '''
    fixes = {'Racket0.d': 'Ball.d', 
             'Racket1.d': 'Ball.d'}
    sort_maps = [('w_res0', 'w_res1'), ('Racket0', 'Racket1')]

    extra_optimization_maps = [
        lambda results: {'FS': 1 - results['w_ball.w']},
        lambda results: {'FS.std': results['w_ball.w.std']}
    ]
    extra_sampling_maps = [
        lambda samples: {
            'FS': np.mean(samples['w_res0.w'] + samples['w_res1.w'], axis=1),
            'FS.std': np.std(samples['w_res0.w'] + samples['w_res1.w'], axis=1)}
    ]

    extra_prior = 'return w_res1.w < w_res0.w;'



from mdt import CompartmentTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDI_EC(CompartmentTemplate):

    parameters = ('g', 'b', 'd', 'dperp0', 'theta', 'phi', 'kappa')
    dependencies = ('dawson', 'Zeppelin')
    cl_code = '''
        mot_float_type tmp;
        mot_float_type dw_0, dw_1;
    
        if(kappa > 1e-5){
            tmp = sqrt(kappa)/dawson(sqrt(kappa));
            dw_0 = ( -(d - dperp0) + 2 * dperp0 * kappa + (d - dperp0) * tmp) / (2.0 * kappa);
            dw_1 = ( (d - dperp0) + 2 * (d+dperp0) * kappa - (d - dperp0) * tmp) / (4.0 * kappa);
        }
        else{
            tmp = 2 * (d - dperp0) * kappa;
            dw_0 = ((2 * dperp0 + d) / 3.0) + (tmp/22.5) + ((tmp * kappa) / 236.0);
            dw_1 = ((2 * dperp0 + d) / 3.0) - (tmp/45.0) - ((tmp * kappa) / 472.0);
        }
    
        return Zeppelin(g, b, dw_0, dw_1, theta, phi);
    '''



__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractParameterDependency(object):

    @property
    def pre_transform_code(self):
        """Some code that may be prefixed to this parameter dependency.

        Here one can put more elaborate CL code. Please make sure that additional variables are unique.

        Returns:
            str: The pre transformation code. This is prepended to the dependency function.
        """
        return ''

    @property
    def assignment_code(self):
        """Get the assignment code (including a ;).

        Returns:
            str: The assignment code.
        """
        return ''


class SimpleAssignment(AbstractParameterDependency):

    def __init__(self, assignment_code):
        """Adds a simple parameter dependency rule for the given parameter.

        This is for one parameter, a simple one-line transformation dependency.

        Args:
            assignment_code (str): the assignment code (in CL) for this parameter
        """
        self._assignment = assignment_code

    @property
    def assignment_code(self):
        return self._assignment



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""This script prints a list of all the models MDT can find in your home directory."""
import argparse
import mdt
import mdt.utils
from mdt.lib.shell_utils import BasicShellApplication

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ListModels(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = __doc__

        parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('-l', '--long', action='store_true', help='print the descriptions')
        return parser

    def run(self, args, extra_args):
        meta_info = mdt.get_models_meta_info()
        models = mdt.get_models_list()

        max_model_name = max(map(len, models))

        for model in models:
            if args.long:
                print(('%-' + str(max_model_name + 2) + 's%-s') % (model, meta_info[model]['description']))
            else:
                print(model)


def get_doc_arg_parser():
    return ListModels().get_documentation_arg_parser()


if __name__ == '__main__':
    ListModels().start()



from mdt import BatchProfileTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-03-18'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class DirPerSubject(BatchProfileTemplate):
    description = '''
        General layout for batch fitting with a folder per subject.
    
        Every subject has its own directory. For every type of file (protocol, bvec, bval, TE, Delta, delta, DWI, mask)
        we use the first one found. 
        
        The pattern matching of the DWI data takes into account filtering the niftis that were already matched by the
        mask and gradient deviations file.

        You can provide the noise standard deviation to use using a noise_std file containing a single float.
        
        The optional items TE, Delta and delta provide extra information about the protocol.
        They should either contain exactly 1 value (for all protocol lines), or a value per protocol line.
        
        Example directory layout:
            /a/data.nii(.gz)
            /a/bval
            /a/bvec

            /b/...
            /c/...

        Optional items:
            /*/prtcl
            /*/*.prtcl

            /*/TE (in seconds)
            /*/delta (in seconds)
            /*/Delta (in seconds) (it is also possible to use a filename /*/big_delta for case insensitive filesystems)

            /*/*_mask.nii(.gz)

            /*/noise_std.{txt,nii,nii.gz}
    '''



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'optimization_options_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_OptimizationOptionsDialog(object):
    def setupUi(self, OptimizationOptionsDialog):
        OptimizationOptionsDialog.setObjectName("OptimizationOptionsDialog")
        OptimizationOptionsDialog.resize(843, 337)
        self.verticalLayout = QtWidgets.QVBoxLayout(OptimizationOptionsDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(OptimizationOptionsDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(OptimizationOptionsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.line = QtWidgets.QFrame(OptimizationOptionsDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_12 = QtWidgets.QLabel(OptimizationOptionsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_12.setFont(font)
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 4, 2, 1, 1)
        self.label = QtWidgets.QLabel(OptimizationOptionsDialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 7, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.defaultOptimizer_True = QtWidgets.QRadioButton(OptimizationOptionsDialog)
        self.defaultOptimizer_True.setChecked(True)
        self.defaultOptimizer_True.setObjectName("defaultOptimizer_True")
        self.defaultOptimizerGroup = QtWidgets.QButtonGroup(OptimizationOptionsDialog)
        self.defaultOptimizerGroup.setObjectName("defaultOptimizerGroup")
        self.defaultOptimizerGroup.addButton(self.defaultOptimizer_True)
        self.horizontalLayout.addWidget(self.defaultOptimizer_True)
        self.defaultOptimizer_False = QtWidgets.QRadioButton(OptimizationOptionsDialog)
        self.defaultOptimizer_False.setObjectName("defaultOptimizer_False")
        self.defaultOptimizerGroup.addButton(self.defaultOptimizer_False)
        self.horizontalLayout.addWidget(self.defaultOptimizer_False)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 1, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.floatPrecision = QtWidgets.QRadioButton(OptimizationOptionsDialog)
        self.floatPrecision.setChecked(True)
        self.floatPrecision.setObjectName("floatPrecision")
        self.floatPrecisionGroup = QtWidgets.QButtonGroup(OptimizationOptionsDialog)
        self.floatPrecisionGroup.setObjectName("floatPrecisionGroup")
        self.floatPrecisionGroup.addButton(self.floatPrecision)
        self.horizontalLayout_2.addWidget(self.floatPrecision)
        self.doublePrecision = QtWidgets.QRadioButton(OptimizationOptionsDialog)
        self.doublePrecision.setObjectName("doublePrecision")
        self.floatPrecisionGroup.addButton(self.doublePrecision)
        self.horizontalLayout_2.addWidget(self.doublePrecision)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.gridLayout.addLayout(self.horizontalLayout_2, 0, 1, 1, 1)
        self.label_13 = QtWidgets.QLabel(OptimizationOptionsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_13.setFont(font)
        self.label_13.setObjectName("label_13")
        self.gridLayout.addWidget(self.label_13, 7, 2, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.recalculateAll_True = QtWidgets.QRadioButton(OptimizationOptionsDialog)
        self.recalculateAll_True.setObjectName("recalculateAll_True")
        self.recalculateAllGroup = QtWidgets.QButtonGroup(OptimizationOptionsDialog)
        self.recalculateAllGroup.setObjectName("recalculateAllGroup")
        self.recalculateAllGroup.addButton(self.recalculateAll_True)
        self.horizontalLayout_3.addWidget(self.recalculateAll_True)
        self.recalculateAll_False = QtWidgets.QRadioButton(OptimizationOptionsDialog)
        self.recalculateAll_False.setChecked(True)
        self.recalculateAll_False.setObjectName("recalculateAll_False")
        self.recalculateAllGroup.addButton(self.recalculateAll_False)
        self.horizontalLayout_3.addWidget(self.recalculateAll_False)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.gridLayout.addLayout(self.horizontalLayout_3, 7, 1, 1, 1)
        self.label_15 = QtWidgets.QLabel(OptimizationOptionsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_15.setFont(font)
        self.label_15.setObjectName("label_15")
        self.gridLayout.addWidget(self.label_15, 0, 2, 1, 1)
        self.patience = QtWidgets.QLineEdit(OptimizationOptionsDialog)
        self.patience.setEnabled(False)
        self.patience.setObjectName("patience")
        self.gridLayout.addWidget(self.patience, 4, 1, 1, 1)
        self.line_4 = QtWidgets.QFrame(OptimizationOptionsDialog)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout.addWidget(self.line_4, 6, 0, 1, 3)
        self.label_11 = QtWidgets.QLabel(OptimizationOptionsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_11.setFont(font)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 3, 2, 1, 1)
        self.optimizationRoutine = QtWidgets.QComboBox(OptimizationOptionsDialog)
        self.optimizationRoutine.setEnabled(False)
        self.optimizationRoutine.setEditable(False)
        self.optimizationRoutine.setObjectName("optimizationRoutine")
        self.gridLayout.addWidget(self.optimizationRoutine, 3, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(OptimizationOptionsDialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1)
        self.label_5 = QtWidgets.QLabel(OptimizationOptionsDialog)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 3, 0, 1, 1)
        self.label_6 = QtWidgets.QLabel(OptimizationOptionsDialog)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 4, 0, 1, 1)
        self.label_8 = QtWidgets.QLabel(OptimizationOptionsDialog)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 0, 0, 1, 1)
        self.label_10 = QtWidgets.QLabel(OptimizationOptionsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 2, 2, 1, 1)
        self.line_5 = QtWidgets.QFrame(OptimizationOptionsDialog)
        self.line_5.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout.addWidget(self.line_5, 1, 0, 1, 3)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem3)
        self.line_3 = QtWidgets.QFrame(OptimizationOptionsDialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.buttonBox = QtWidgets.QDialogButtonBox(OptimizationOptionsDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(OptimizationOptionsDialog)
        self.buttonBox.accepted.connect(OptimizationOptionsDialog.accept)
        self.buttonBox.rejected.connect(OptimizationOptionsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(OptimizationOptionsDialog)
        OptimizationOptionsDialog.setTabOrder(self.floatPrecision, self.doublePrecision)
        OptimizationOptionsDialog.setTabOrder(self.doublePrecision, self.defaultOptimizer_True)
        OptimizationOptionsDialog.setTabOrder(self.defaultOptimizer_True, self.defaultOptimizer_False)
        OptimizationOptionsDialog.setTabOrder(self.defaultOptimizer_False, self.optimizationRoutine)
        OptimizationOptionsDialog.setTabOrder(self.optimizationRoutine, self.patience)
        OptimizationOptionsDialog.setTabOrder(self.patience, self.recalculateAll_True)
        OptimizationOptionsDialog.setTabOrder(self.recalculateAll_True, self.recalculateAll_False)

    def retranslateUi(self, OptimizationOptionsDialog):
        _translate = QtCore.QCoreApplication.translate
        OptimizationOptionsDialog.setWindowTitle(_translate("OptimizationOptionsDialog", "Optimization options"))
        self.label_3.setText(_translate("OptimizationOptionsDialog", "Optimization options"))
        self.label_4.setText(_translate("OptimizationOptionsDialog", "Advanced options for the model fitting procedure"))
        self.label_12.setText(_translate("OptimizationOptionsDialog", "(Scales the number of iterations)"))
        self.label.setText(_translate("OptimizationOptionsDialog", "Recalculate all:"))
        self.defaultOptimizer_True.setText(_translate("OptimizationOptionsDialog", "Yes   "))
        self.defaultOptimizer_False.setText(_translate("OptimizationOptionsDialog", "No"))
        self.floatPrecision.setText(_translate("OptimizationOptionsDialog", "Float"))
        self.doublePrecision.setText(_translate("OptimizationOptionsDialog", "Double"))
        self.label_13.setText(_translate("OptimizationOptionsDialog", "(For cascades, if we want to recalculate the entire chain)"))
        self.recalculateAll_True.setText(_translate("OptimizationOptionsDialog", "Yes   "))
        self.recalculateAll_False.setText(_translate("OptimizationOptionsDialog", "No"))
        self.label_15.setText(_translate("OptimizationOptionsDialog", "(The precision for the calculations)"))
        self.label_11.setText(_translate("OptimizationOptionsDialog", "(Manual select the routine to use)"))
        self.label_2.setText(_translate("OptimizationOptionsDialog", "Use default optimizer:"))
        self.label_5.setText(_translate("OptimizationOptionsDialog", "Optimization routine:"))
        self.label_6.setText(_translate("OptimizationOptionsDialog", "Patience:"))
        self.label_8.setText(_translate("OptimizationOptionsDialog", "Float precision:"))
        self.label_10.setText(_translate("OptimizationOptionsDialog", "(Enables manual selection of the optimization routine)"))

